import traceback
from copy import deepcopy
from datetime import datetime as dt
from time import sleep, time

from naeural_core import DecentrAIObject, Logger
from naeural_core import constants as ct
from naeural_core.data_structures import MetadataObject
from naeural_core.local_libraries import _ConfigHandlerMixin
from naeural_core.utils.plugins_base.bc_wrapper import BCWrapper
from naeural_core.utils.shm_manager import SharedMemoryManager

_CONFIG = {
  'IS_THREAD'               : False,
  'URL'                     : None,
  
  'MAX_IDLE_TIME'           : 10,
  
  'STREAM_CONFIG_METADATA'  : {},
  
  'PIPELINE_COMMAND'        : None,
  
  'SESSION_ID'              : None,
  'INITIATOR_ID'            : None,
  
  'USE_LOCAL_COMMS_ONLY'     : False,
  
  'VALIDATION_RULES' : {
    'IS_THREAD' : {'TYPE' : 'bool'},
  },
}

class BaseDataCapture(DecentrAIObject, _ConfigHandlerMixin):
  CONFIG = _CONFIG
  def __init__(
    self, 
    log, 
    default_config, 
    upstream_config, 
    environment_variables, 
    shmem, 
    signature, 
    fn_loop_stage_callback, 
    **kwargs
  ):
    self._default_config = default_config
    self._upstream_config = upstream_config
    self._environment_variables = environment_variables

    self.shmem = shmem
    self.__blockchain_manager = shmem[ct.BLOCKCHAIN_MANAGER]
    
    self.bc = BCWrapper(self.__blockchain_manager, owner=self) # blockchain wrapper (for encryption/decryption
    
    self.shm_manager = None
    self.nr_connection_issues = 0
    self._fn_loop_stage_callback = fn_loop_stage_callback
    self._initial_setup_done = False
    
    self.config = None
    self._signature = signature
    self._stream_metadata = MetadataObject() # this is a overall metadata object
    
    self.__last_data_timestamp = time()
    
    self.__require_on_config_trigger = False # on_config event trigger

    # control variables used mostly in DCT but declared here for functions defined in BaseDataCapture
    self._capture_loop_in_exec = False
    self._loop_paused = False
    self._capture_thread_waiting = False
    # end loop control variables
    
    super(BaseDataCapture, self).__init__(log=log, **kwargs)
    return
  
  def P(self, s, color=None, **kwargs):
    if color is None or (isinstance(color,str) and color[0] not in ['e', 'r']):
      color = ct.COLORS.DCT
    super().P(s, prefix=True, color=color, **kwargs)
    return
  
  
  @property 
  def is_idle_alert(self):
    if self.cfg_is_thread:
      # only works for DCTs
      margin = 1 / self.cfg_cap_resolution
    else:
      margin = 1

    is_idle = self.time_since_last_data > (margin + self.cfg_max_idle_time)
    return is_idle  
  
  
  @property
  def in_initial_setup(self):
    return not self._initial_setup_done
  
  @property
  def time_since_last_input(self):
    return -1 # defined only in DCT
  
  @property
  def time_since_last_data(self):
    elapsed = time() - self.__last_data_timestamp
    return elapsed
    

  def startup(self):
    super().startup()
    self._update_config()
    self.shm_manager = SharedMemoryManager(
      dct_shared=self.shmem,
      stream=self.cfg_name,
      plugin='DEFAULT',
      instance='DEFAULT',
      category=ct.SHMEM.DCT,
      linked_instances=None,
      log=self.log,
    )
    self._populate_stream_metadata_from_config()
    return
  
  def _populate_stream_metadata_from_config(self):
    """
    This protected method populates the `_stream_metadata` object with basic information.
    Should be populated in derived classes.

    Returns
    -------
    None.

    """
    return
  
  

  @property
  def cfg_alive_until(self):
    alive_until = self.cfg_stream_config_metadata.get('ALIVE_UNTIL', None)
    if alive_until is not None:
      alive_until = dt.strptime(alive_until, '%Y-%m-%d %H:%M')
    return alive_until


  @property
  def alive_until_passed(self):
    alive_until = self.cfg_alive_until
    if alive_until is not None:
      return (alive_until - dt.now()).total_seconds() < 0
    return False


  @property
  def has_data(self):
    return False


  @property
  def can_delete(self):
    return False


  @property
  def _device_id(self):
    eeid = self.log.config_data.get(ct.CONFIG_STARTUP_v2.K_EE_ID, '')[:ct.EE_ALIAS_MAX_SIZE]
    return eeid



  def _create_notification(self, notif, msg, notif_code=None, info=None, stream_name=None, error_code=None, **kwargs):
    return super()._create_notification(
      notif=notif, 
      notif_code=notif_code,
      msg=msg, 
      info=info,
      stream_name=self.cfg_name,
      session_id=self.cfg_session_id,
      initiator_id=self.cfg_initiator_id,
      error_code=error_code,
      ct=ct,
      use_local_comms_only=self.cfg_use_local_comms_only,
      **kwargs
    )


  def get_nr_parallel_captures(self):
    return self.shmem[ct.CAPTURE_MANAGER][ct.NR_CAPTURES]
  
  
  def maybe_trigger_on_config(self):
    if self.__require_on_config_trigger:
      self.__on_config_changed()
      self.__require_on_config_trigger = False
    return

  def get_plugin_specific_stats(self):
    """
    This method should be implemented in derived classes if they need to return user defined stats

    Returns
    -------
    status : dict
      The user defined status
    """
    result = {
    }
    return result

  def maybe_update_config(self, upstream_config):
    # if the following keys are modified we do not need to update the config
    EXCEPT_KEYS = [
      ct.CONFIG_STREAM.PLUGINS,
      ct.CONFIG_STREAM.LAST_UPDATE_TIME,
      ct.CONFIG_STREAM.K_MODIFIED_BY_ADDR,
      ct.CONFIG_STREAM.K_MODIFIED_BY_ID, 
    ]
    
    if upstream_config == self._upstream_config:
      return

    self.P("<MAIN THR> Changing config for {}".format(self), color='b')
    self._upstream_config = upstream_config
    
    need_update, lst_to_update = self.needs_update(self._upstream_config, except_keys=EXCEPT_KEYS)
    if not need_update:
      self.P("  <MAIN THR> No need to update config for {}. Modified {}, Excluded {}".format(
        self, lst_to_update, EXCEPT_KEYS), color='b'
      )
      self.config_data[ct.CONFIG_STREAM.LAST_UPDATE_TIME] = self._upstream_config.get(ct.CONFIG_STREAM.LAST_UPDATE_TIME, None)
      return
    
    # now while changing config we must stop loop exec
    self._fn_loop_stage_callback('4.collect.update_streams._check_captures.maybe_update_config.{}.wait'.format(self.cfg_name))
    # while self._capture_loop_in_exec:
    #   sleep(0.001)
    self._loop_paused = True
    start_waiting = time()
    while not self._capture_thread_waiting:
      sleep(0.001)
      if (time() - start_waiting) > 2: # 3s
        #TODO(AID): send notification with this behavior\
        self.P("<MAIN THR> WARNING! Updating through stream collecting!", color='r')
        break
    self._fn_loop_stage_callback('4.collect.update_streams._check_captures.maybe_update_config.{}.post-wait'.format(self.cfg_name))
    
    
    last_config = deepcopy(self.config_data)
    self._fn_loop_stage_callback('4.collect.update_streams._check_captures.maybe_update_config.{}.post-deepcopy'.format(self.cfg_name))
    try:
      self._update_config()
      self._fn_loop_stage_callback('4.collect.update_streams._check_captures.maybe_update_config.{}.post-update'.format(self.cfg_name))
      # resume loop
      self._loop_paused = False
      # TODO: add specific config modifications to below message
      msg = "Successfully updated data capture thread config for pipeline '{}' with new config".format(self.cfg_name)
      self.P("<MAIN THR> " + msg)
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_NORMAL, 
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_DCT_CONFIG_OK,
        msg=msg,
        displayed=True,
      )
    except Exception as exc:
      # rollback
      self.config = last_config
      self.config_data = last_config
      msg = "Exception occured while updating the capture config: '{}' Rollback to the last good config.".format(exc)
      info = traceback.format_exc()
      self.P(msg, color='r')

      self.__save_config(keys=[k for k in upstream_config.keys() if k in self.config])

      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION, 
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_DCT_CONFIG_FAILED,
        msg=msg,
        info=info,
        displayed=True,
      )

      # resume loop even on exception
      self._loop_paused = False

    self.__require_on_config_trigger = True # TODO: WARNING: simple plugins update trigger reconnects!!!
    self._fn_loop_stage_callback('4.collect.update_streams._check_captures.maybe_update_config.{}.post-callback'.format(self.cfg_name))
    return
  
  
  def _on_config_changed(self):
    """
    This should be defined in each individual DCT to act upon changes - i.e. such as the case when a
    video stream DCT receives a new URL and needs to reconnect
    """
    return
  
  def __on_config_changed(self):
    self.P("********************************************************")
    self.P("On config triggered. Current data connection issues: {}".format(
      self.nr_connection_issues))
    self.P("********************************************************")
    self._on_config_changed()
    return


  def _maybe_add_mandatory_plugin_to_pipeline(self, dct_config):
    """
    Add default mandatory plugins to pipeline, if the plugins do not already exist
    and if pipeline requires them.

    Parameters
    ----------
    dct_config : dict
      the pipeline config.

    Returns
    -------
    None.

    """
    dct_default_signature_instances = dct_config.get(ct.CONFIG_STREAM.DEFAULT_PLUGINS, {})
    dct_default_signature_instances = {
      signature: [
        {
          **instance,
          ct.INSTANCE_ID: instance.get(ct.INSTANCE_ID, self.log.get_short_name(signature) + '_default_' + str(i))
        }
        for i, instance in enumerate(instances)
      ]
      for signature, instances in dct_default_signature_instances.items()
    }

    plugins = dct_config.get(ct.CONFIG_STREAM.PLUGINS, [])
    dct_existing_signature_instances = {x[ct.SIGNATURE] : x[ct.INSTANCES] for x in plugins}
    lst_added_signature_instance_ids = []

    for default_plugin_signature, default_plugin_instances in dct_default_signature_instances.items():
      if default_plugin_signature not in dct_existing_signature_instances:
        dct_existing_signature_instances[default_plugin_signature] = default_plugin_instances
        plugins.append({ ct.SIGNATURE : default_plugin_signature, ct.INSTANCES : default_plugin_instances })
        lst_added_signature_instance_ids.extend([(default_plugin_signature, instance[ct.INSTANCE_ID]) for instance in default_plugin_instances])
      else:
        for default_instance_config in default_plugin_instances:
          if default_instance_config[ct.INSTANCE_ID] not in [x[ct.INSTANCE_ID] for x in dct_existing_signature_instances[default_plugin_signature]]:
            dct_existing_signature_instances[default_plugin_signature].append(default_instance_config)
            lst_added_signature_instance_ids.append((default_plugin_signature, default_instance_config[ct.INSTANCE_ID]))
          else:
            overwrite_config = dct_config.get(ct.CONFIG_STREAM.OVERWRITE_DEFAULT_PLUGIN_CONFIG, False)
            if overwrite_config:
              pos, instance_config = next((pos, x) for pos, x in enumerate(dct_existing_signature_instances[default_plugin_signature]) if x[ct.INSTANCE_ID] == default_instance_config[ct.INSTANCE_ID])
              if instance_config != {**instance_config, **default_instance_config}:
                dct_existing_signature_instances[default_plugin_signature][pos].update(default_instance_config)
                lst_added_signature_instance_ids.append((default_plugin_signature, default_instance_config[ct.INSTANCE_ID]))
              # endif default config different from existing config
            # endif overwrite config
          # endif instance not in default instances
        # end for default instances
      # endif signature not in plugins
    # end for default plugins

    if len(lst_added_signature_instance_ids) > 0:
      self.P("Adding mandatory plugins {} to pipeline '{}'".format(
        ", ".join("'{}:{}'".format(sig, inst) for sig, inst in lst_added_signature_instance_ids), 
        dct_config[ct.NAME]
      ))
      self.__save_config(keys=[ct.CONFIG_STREAM.PLUGINS])
    return

  def _update_config(self):
    # special section for PIPELINE_COMMAND
    is_pipeline_command = False 
    if self.config is not None:
      current_pipeline_command = self.config.get('PIPELINE_COMMAND', None)
      new_pipeline_command = self._upstream_config.get('PIPELINE_COMMAND', None)
      if new_pipeline_command is not None and current_pipeline_command != new_pipeline_command:
        is_pipeline_command = True
    # end special section
    
    self.__name__ = 'DCT:' + self.log.name_abbreviation(self._signature)
    if self._environment_variables is not None and len(self._environment_variables) > 0:
      self.P("<MAIN THR> Updating capture config with node environment config....", color='m')
      self.config = self._merge_prepare_config(
        default_config=self.config,
        delta_config=self._environment_variables,
      )
    #endif

    # next line will create a new config basically so we need then to validate and re-create the cfg_* handlers
    # otherwise they will point to the old data
    self.config = self._merge_prepare_config(default_config=self.config)    

    # now we set config_data, create handlers and validate
    self.setup_config_and_validate(dct_config=self.config)

    # now we have cfg_name and we can use it
    self.__name__ = 'DCT:' + self.log.name_abbreviation(self._signature) + '=' + self.cfg_name
    self._stream_metadata.update(**self.cfg_stream_config_metadata)

    # now we add mandatory plugins to the pipeline
    # these plugins are defined under "DEFAULT_PLUGINS"
    self._maybe_add_mandatory_plugin_to_pipeline(dct_config=self.config)

    self._initial_setup_done = True
    return

  def data_template(self, inputs=None, metadata=None):
    if inputs is None:
      inputs = []

    if metadata is None:
      metadata = self._stream_metadata.__dict__
    
    self.__last_data_timestamp = time()
    
    return {
      'STREAM_NAME' : self.cfg_name, ##TODO ct
      'STREAM_METADATA' : metadata.copy(), ## We copy it because we need to know the "snapshot" metadata for each append TODO ct
      'INPUTS' : inputs ##TODO ct
    }

  def stop(self, join_time=10):
    return

  def __save_config(self, keys):
    # writes the config for a set of particular keys on the local cache!
    save_config_fn = self.shmem[ct.CALLBACKS.PIPELINE_CONFIG_SAVER_CALLBACK]
    try:
      config = {k: self.config[k] for k in keys}
      save_config_fn(self.cfg_name, config_data=config)
      # the save will trigger a update config when the BusinessManager will get the new config from
      # the ConfigManager and will pass it as upstream_config to the plugin
      # so we already update the upstream_config to avoid this
      for k in keys:
        self._upstream_config[k] = self.config[k]
    except Exception as exc:
      self.P("Error '{}' while saving keys {}".format(exc, keys))
      raise exc
    return

  
  def __repr__(self):
    return '<{}:{}>'.format(self._signature, self.cfg_name)



import os
import torch as th
import torchvision as tv
from ultralytics import YOLO


def xywh2xyxy(x):
  """
  Convert bounding box coordinates from (x, y, width, height) format to (x1, y1, x2, y2) format where (x1, y1) is the
  top-left corner and (x2, y2) is the bottom-right corner.

  Args:
      x (np.ndarray) or (torch.Tensor): The input bounding box coordinates in (x, y, width, height) format.
  Returns:
      y (np.ndarray) or (torch.Tensor): The bounding box coordinates in (x1, y1, x2, y2) format.
  """
  y = x.clone()
  y[..., 0] = x[..., 0] - x[..., 2] / 2  # top left x
  y[..., 1] = x[..., 1] - x[..., 3] / 2  # top left y
  y[..., 2] = x[..., 0] + x[..., 2] / 2  # bottom right x
  y[..., 3] = x[..., 1] + x[..., 3] / 2  # bottom right y
  return y


@th.jit.script
def y8_nms(
    prediction,
    # conf_thres:float=0.25,
    # iou_thres:float=0.45,
    # classes=None,
    # agnostic=False,
    # multi_label:bool=False,
    # max_det:int=300,
    # max_time_img:float=0.05,
    # max_nms:int=30000,
    # max_wh:int=7680,
):
  """
  Perform non-maximum suppression (NMS) on a set of boxes, with support for masks and multiple labels per box.

  Arguments:
      prediction (th.Tensor): A tensor of shape (batch_size, num_boxes, num_classes + 4 + num_masks)
          containing the predicted boxes, classes, and masks. The tensor should be in the format
          output by a model, such as YOLO.
      conf_thres (float): The confidence threshold below which boxes will be filtered out.
          Valid values are between 0.0 and 1.0.
      iou_thres (float): The IoU threshold below which boxes will be filtered out during NMS.
          Valid values are between 0.0 and 1.0.
      classes (List[int]): A list of class indices to consider. If None, all classes will be considered.
      agnostic (bool): If True, the model is agnostic to the number of classes, and all
          classes will be considered as one.
      multi_label (bool): If True, each box may have multiple labels.
      labels (List[List[Union[int, float, th.Tensor]]]): A list of lists, where each inner
          list contains the apriori labels for a given image. The list should be in the format
          output by a dataloader, with each label being a tuple of (class_index, x1, y1, x2, y2).
      max_det (int): The maximum number of boxes to keep after NMS.
      nc (int): (optional) The number of classes output by the model. Any indices after this will be considered masks.
      max_time_img (float): The maximum time (seconds) for processing one image.
      max_nms (int): The maximum number of boxes into tv.ops.nms().
      max_wh (int): The maximum box width and height in pixels

  Returns:
      (List[th.Tensor]): A list of length batch_size, where each element is a tensor of
          shape (num_boxes, 6 + num_masks) containing the kept boxes, with columns
          (x1, y1, x2, y2, confidence, class, mask1, mask2, ...).
  """

  bs = prediction.shape[0]  # batch size
  nc = prediction.shape[1] - 4  # number of classes
  nm = 0  # prediction.shape[1] - nc - 4
  mi = 4 + nc  # mask start index
  xc = prediction[:, 4:mi].amax(1) > 0.25  # candidates  #conf_thres

  th_output = th.zeros((bs, 300, 6 + nm), device=prediction.device)
  th_n_det = th.zeros(bs, device=prediction.device, dtype=th.int32)

  for xi in range(prediction.shape[0]):  # image index, image inference
    x = prediction[xi]
    x = x.transpose(0, -1)[xc[xi]]  # confidence

    # If none remain process next image
    if x.shape[0] > 0:

      # Detections matrix nx6 (xyxy, conf, cls)
      box, cls, mask = x.split((4, nc, nm), 1)
      box = xywh2xyxy(box)  # center_x, center_y, width, height) to (x1, y1, x2, y2)

      conf, j = cls.max(1, keepdim=True)
      x = th.cat((box, conf, j.float(), mask), 1)

      # Check shape
      n = x.shape[0]  # number of boxes
      if n > 0:  # no boxes
        x = x[x[:, 4].argsort(descending=True)[:30000]]  # sort by confidence and remove excess boxes #max_nms

        # Batched NMS
        c = x[:, 5:6] * 7680  # classes #max_wh
        boxes = x[:, :4] + c
        scores = x[:, 4]  # boxes (offset by class), scores
        i = tv.ops.nms(boxes, scores, 0.45)  # NMS #iou_thres
        i = i[:300]  # limit detections #max_det

        th_n_det[xi] = i.shape[0]
        th_output[xi, :th_n_det[xi], :] = x[i]

  return th_output, th_n_det


class BaseYoloExtended(th.nn.Module):
  def __init__(self, base_model, **kwargs):
    super().__init__()
    self.backbone = base_model
    self.y8_nms = y8_nms
    return

  def forward(self, inputs):
    th_x = self.backbone_model(inputs)
    th_x = self.y8_nms(th_x)
    return th_x


class YoloExtended(YOLO):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return

  def convert_model(self):
    self.model = BaseYoloExtended(self.model)
    return


if __name__ == '__main__':
  pass
  # load images
  img_dir = r'C:\Users\bleot\Dropbox\DATA\_vapor_data\__tests\Y8'
  img_paths = [
    os.path.join(img_dir, img_name) for img_name in os.listdir(img_dir) if img_name.endswith(('.jpg', 'png'))
  ]

  # preprocess images

  # load model

  # convert model

  # for format in formats export in format

  # for format in formats load model

  # for format in formats warmup model

  # for format in formats benchmark model





"""
IMPORTANT:
  The command handlers are defined in this file. The actual commands are dinamically inferred from the available handlers.
"""

from naeural_core import constants as ct


class ExecutionEngineCommandHandlers:
    
  ###########################################
  #         command definition area         #
  ###########################################
  
  def get_commands_list(self):
    return list(self.get_cmd_handlers().keys())
  
  
  def cmd_handler_stop(self, initiator_id=None, session_id=None, **kwargs):    
    msg = "  Running 'STOP' command in main loop received from '{}'".format(initiator_id)
    self.P(msg, color='y')
    self._create_notification(
      notif=ct.NOTIFICATION_TYPE.STATUS_NORMAL,
      msg=msg,
      initiator_id=initiator_id,
      session_id=session_id,
    )    
    return ct.CODE_STOP, ct.DEVICE_STATUS_STOP
  
  def cmd_handler_restart(self, initiator_id=None, session_id=None, **kwargs):
    msg = "  Running 'RESTART' command in main loop received from '{}'".format(initiator_id)
    self.P(msg, color='y')
    self._create_notification(
      notif=ct.NOTIFICATION_TYPE.STATUS_NORMAL,
      msg=msg,
      initiator_id=initiator_id,
      session_id=session_id,
    )    
    return ct.CODE_RESTART, ct.DEVICE_STATUS_RESTART

  def cmd_handler_pipeline_command(self, command_content=None, initiator_id=None, session_id=None, **kwargs):
    self.P("  Running 'PIPELINE_COMMAND' command in main loop received from '{}'".format(initiator_id), color='y')
    self._config_manager.pipeline_command(
      dct_config_data=command_content, 
      initiator_id=initiator_id, session_id=session_id, 
      **kwargs
    )
    return

  def cmd_handler_update_pipeline_instance(self, command_content=None, initiator_id=None, session_id=None, **kwargs):
    self.P("  Running 'UPDATE_PIPELINE_INSTANCE' command in main loop received from '{}'".format(initiator_id), color='y')
    self._config_manager.update_pipeline_instance(
      dct_config_data=command_content,
      initiator_id=initiator_id, session_id=session_id
    )
    return
  
  def cmd_handler_batch_update_pipeline_instance(self, command_content=None, initiator_id=None, session_id=None, **kwargs):
    self.P("  Running 'BATCH_UPDATE_PIPELINE_INSTANCE' command in main loop received from '{}'".format(initiator_id), color='y')
    self._config_manager.batch_update_pipeline_instance(
      lst_configs=command_content,
      initiator_id=initiator_id, session_id=session_id,
    )
    return
  
  
  def cmd_handler_update_config(self, command_content=None, initiator_id=None, session_id=None, **kwargs):
    self.P("  Running 'UPDATE_CONFIG' command in main loop received from '{}'".format(initiator_id), color='y')
    self._config_manager.update_config_stream(
      delta_config_stream=command_content, 
      initiator_id=initiator_id, session_id=session_id
    )
    return

  def cmd_handler_delete_config(self, command_content=None, initiator_id=None, session_id=None, **kwargs):
    self.P("  Running 'DELETE_CONFIG' command in main loop received from '{}'".format(initiator_id), color='y')
    self._config_manager.delete_config_stream(
      stream_name=command_content, 
      initiator_id=initiator_id, session_id=session_id
    )
    return

  def cmd_handler_delete_config_all(self, command_content=None, initiator_id=None, session_id=None, **kwargs):
    msg = "  Running 'DELETE_CONFIG_ALL' command in main loop received from '{}'".format(initiator_id)
    self.P(msg, color='y')
    self._create_notification(
      notif=ct.NOTIFICATION_TYPE.STATUS_NORMAL,
      msg=msg,
      initiator_id=initiator_id,
    )    
    self._config_manager.delete_all_streams(initiator_id=initiator_id, session_id=session_id)
    return
      
  
  def cmd_handler_archive_config(self, command_content=None, initiator_id=None, session_id=None, **kwargs):
    self.P("  Running 'ARCHIVE_CONFIG' command in main loop received from '{}'".format(initiator_id), color='y')
    self._config_manager.archive_single_stream(stream_name=command_content, initiator_id=initiator_id, session_id=session_id)
    return


  def cmd_handler_archive_config_all(self, command_content=None, initiator_id=None, session_id=None, **kwargs):
    msg = "  Running 'ARCHIVE_CONFIG_ALL' command in main loop received from '{}'".format(initiator_id)
    self.P(msg, color='y')
    self._create_notification(
      notif=ct.NOTIFICATION_TYPE.STATUS_NORMAL,
      msg=msg,
      initiator_id=initiator_id,
    )    
    self._config_manager.archive_all_streams(initiator_id=initiator_id, session_id=session_id)
    return


  def cmd_handler_reload_config_from_disk(self, initiator_id=None, session_id=None, **kwargs):
    msg = "  Running 'RELOAD_CONFIG_FROM_DISK' command in main loop received from '{}'".format(initiator_id)
    self.P(msg, color='y')
    self._create_notification(
      notif=ct.NOTIFICATION_TYPE.STATUS_NORMAL,
      msg=msg,
      initiator_id=initiator_id,
    )    
    self._config_manager.load_streams_configurations()
    return


  def cmd_handler_simple_heartbeat(self, initiator_id=None, session_id=None, **kwargs):
    self.P("  Running 'SIMPLE_HEARTBEAT' command in main loop coming from '{}'".format(initiator_id), color='y')
    self._maybe_send_heartbeat(
      status=None,
      full_info=False,
      send_log=False,
      force=True,
      initiator_id=initiator_id,
      session_id=session_id,
    )
    return

  
  def cmd_handler_timers_only_heartbeat(self, initiator_id=None, session_id=None, **kwargs):
    self.P("  Running 'TIMERS_ONLY_HEARTBEAT' command in main loop coming from '{}'".format(initiator_id), color='y')
    self._maybe_send_heartbeat(
      status=None,
      full_info=True,
      send_log=False,
      force=True,
      initiator_id=initiator_id,
      session_id=session_id,
    )
    return

  def cmd_handler_full_heartbeat(self, initiator_id=None, session_id=None, **kwargs):
    self.P("  Running `FULL_HEARTBEAT` command in main loop coming from '{}'".format(initiator_id), color='y')
    self._maybe_send_heartbeat(
      status=None,
      full_info=True,
      send_log=True,
      force=True,
      initiator_id=initiator_id,
      session_id=session_id,
    )
    return

  def cmd_handler_reset_whitelist_commands_to_template(self, initiator_id=None, session_id=None, **kwargs):
    self.P("  Running 'RESET_WHITELIST_COMMANDS_TO_TEMPLATE' command in main loop received from '{}'".format(initiator_id), color='y')
    self.comm_manager._reset_whitelist_commands_to_template()
    return

  ##############################################
  #         end command definition area        #
  ##############################################



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Validate a trained YOLOv5 model accuracy on a custom dataset

Usage:
    $ python path/to/val.py --data coco128.yaml --weights yolov5s.pt --img 640
"""

import argparse
import json
import os
import sys
from pathlib import Path
from threading import Thread

import numpy as np
import torch
from tqdm import tqdm

FILE = Path(__file__).resolve()
ROOT = FILE.parents[0]  # YOLOv5 root directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH
ROOT = ROOT.relative_to(Path.cwd())  # relative

from models.experimental import attempt_load
from utils.datasets import create_dataloader
from utils.general import coco80_to_coco91_class, check_dataset, check_img_size, check_requirements, \
    check_suffix, check_yaml, box_iou, non_max_suppression, scale_coords, xyxy2xywh, xywh2xyxy, set_logging, \
    increment_path, colorstr, print_args
from utils.metrics import ap_per_class, ConfusionMatrix
from utils.plots import output_to_target, plot_images, plot_val_study
from utils.torch_utils import select_device, time_sync
from utils.callbacks import Callbacks


def save_one_txt(predn, save_conf, shape, file):
    # Save one txt result
    gn = torch.tensor(shape)[[1, 0, 1, 0]]  # normalization gain whwh
    for *xyxy, conf, cls in predn.tolist():
        xywh = (xyxy2xywh(torch.tensor(xyxy).view(1, 4)) / gn).view(-1).tolist()  # normalized xywh
        line = (cls, *xywh, conf) if save_conf else (cls, *xywh)  # label format
        with open(file, 'a') as f:
            f.write(('%g ' * len(line)).rstrip() % line + '\n')


def save_one_json(predn, jdict, path, class_map):
    # Save one JSON result {"image_id": 42, "category_id": 18, "bbox": [258.15, 41.29, 348.26, 243.78], "score": 0.236}
    image_id = int(path.stem) if path.stem.isnumeric() else path.stem
    box = xyxy2xywh(predn[:, :4])  # xywh
    box[:, :2] -= box[:, 2:] / 2  # xy center to top-left corner
    for p, b in zip(predn.tolist(), box.tolist()):
        jdict.append({'image_id': image_id,
                      'category_id': class_map[int(p[5])],
                      'bbox': [round(x, 3) for x in b],
                      'score': round(p[4], 5)})


def process_batch(detections, labels, iouv):
    """
    Return correct predictions matrix. Both sets of boxes are in (x1, y1, x2, y2) format.
    Arguments:
        detections (Array[N, 6]), x1, y1, x2, y2, conf, class
        labels (Array[M, 5]), class, x1, y1, x2, y2
    Returns:
        correct (Array[N, 10]), for 10 IoU levels
    """
    correct = torch.zeros(detections.shape[0], iouv.shape[0], dtype=torch.bool, device=iouv.device)
    iou = box_iou(labels[:, 1:], detections[:, :4])
    x = torch.where((iou >= iouv[0]) & (labels[:, 0:1] == detections[:, 5]))  # IoU above threshold and classes match
    if x[0].shape[0]:
        matches = torch.cat((torch.stack(x, 1), iou[x[0], x[1]][:, None]), 1).cpu().numpy()  # [label, detection, iou]
        if x[0].shape[0] > 1:
            matches = matches[matches[:, 2].argsort()[::-1]]
            matches = matches[np.unique(matches[:, 1], return_index=True)[1]]
            # matches = matches[matches[:, 2].argsort()[::-1]]
            matches = matches[np.unique(matches[:, 0], return_index=True)[1]]
        matches = torch.Tensor(matches).to(iouv.device)
        correct[matches[:, 1].long()] = matches[:, 2:3] >= iouv
    return correct


@torch.no_grad()
def run(data,
        weights=None,  # model.pt path(s)
        batch_size=32,  # batch size
        imgsz=640,  # inference size (pixels)
        conf_thres=0.001,  # confidence threshold
        iou_thres=0.6,  # NMS IoU threshold
        task='val',  # train, val, test, speed or study
        device='',  # cuda device, i.e. 0 or 0,1,2,3 or cpu
        single_cls=False,  # treat as single-class dataset
        augment=False,  # augmented inference
        verbose=False,  # verbose output
        save_txt=False,  # save results to *.txt
        save_hybrid=False,  # save label+prediction hybrid results to *.txt
        save_conf=False,  # save confidences in --save-txt labels
        save_json=False,  # save a COCO-JSON results file
        project=ROOT / 'runs/val',  # save to project/name
        name='exp',  # save to project/name
        exist_ok=False,  # existing project/name ok, do not increment
        half=True,  # use FP16 half-precision inference
        model=None,
        dataloader=None,
        save_dir=Path(''),
        plots=True,
        callbacks=Callbacks(),
        compute_loss=None,
        ):
    # Initialize/load model and set device
    training = model is not None
    if training:  # called by train.py
        device = next(model.parameters()).device  # get model device

    else:  # called directly
        device = select_device(device, batch_size=batch_size)

        # Directories
        save_dir = increment_path(Path(project) / name, exist_ok=exist_ok)  # increment run
        (save_dir / 'labels' if save_txt else save_dir).mkdir(parents=True, exist_ok=True)  # make dir

        # Load model
        check_suffix(weights, '.pt')
        model = attempt_load(weights, map_location=device)  # load FP32 model
        gs = max(int(model.stride.max()), 32)  # grid size (max stride)
        imgsz = check_img_size(imgsz, s=gs)  # check image size

        # Multi-GPU disabled, incompatible with .half() https://github.com/ultralytics/yolov5/issues/99
        # if device.type != 'cpu' and torch.cuda.device_count() > 1:
        #     model = nn.DataParallel(model)

        # Data
        data = check_dataset(data)  # check

    # Half
    half &= device.type != 'cpu'  # half precision only supported on CUDA
    model.half() if half else model.float()

    # Configure
    model.eval()
    is_coco = isinstance(data.get('val'), str) and data['val'].endswith('coco/val2017.txt')  # COCO dataset
    nc = 1 if single_cls else int(data['nc'])  # number of classes
    iouv = torch.linspace(0.5, 0.95, 10).to(device)  # iou vector for mAP@0.5:0.95
    niou = iouv.numel()

    # Dataloader
    if not training:
        if device.type != 'cpu':
            model(torch.zeros(1, 3, imgsz, imgsz).to(device).type_as(next(model.parameters())))  # run once
        task = task if task in ('train', 'val', 'test') else 'val'  # path to train/val/test images
        dataloader = create_dataloader(data[task], imgsz, batch_size, gs, single_cls, pad=0.5, rect=True,
                                       prefix=colorstr(f'{task}: '))[0]

    seen = 0
    confusion_matrix = ConfusionMatrix(nc=nc)
    names = {k: v for k, v in enumerate(model.names if hasattr(model, 'names') else model.module.names)}
    class_map = coco80_to_coco91_class() if is_coco else list(range(1000))
    s = ('%20s' + '%11s' * 6) % ('Class', 'Images', 'Labels', 'P', 'R', 'mAP@.5', 'mAP@.5:.95')
    dt, p, r, f1, mp, mr, map50, map = [0.0, 0.0, 0.0], 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
    loss = torch.zeros(3, device=device)
    jdict, stats, ap, ap_class = [], [], [], []
    for batch_i, (img, targets, paths, shapes) in enumerate(tqdm(dataloader, desc=s)):
        t1 = time_sync()
        img = img.to(device, non_blocking=True)
        img = img.half() if half else img.float()  # uint8 to fp16/32
        img /= 255.0  # 0 - 255 to 0.0 - 1.0
        targets = targets.to(device)
        nb, _, height, width = img.shape  # batch size, channels, height, width
        t2 = time_sync()
        dt[0] += t2 - t1

        # Run model
        out, train_out = model(img, augment=augment)  # inference and training outputs
        dt[1] += time_sync() - t2

        # Compute loss
        if compute_loss:
            loss += compute_loss([x.float() for x in train_out], targets)[1]  # box, obj, cls

        # Run NMS
        targets[:, 2:] *= torch.Tensor([width, height, width, height]).to(device)  # to pixels
        lb = [targets[targets[:, 0] == i, 1:] for i in range(nb)] if save_hybrid else []  # for autolabelling
        t3 = time_sync()
        out = non_max_suppression(out, conf_thres, iou_thres, labels=lb, multi_label=True, agnostic=single_cls)
        dt[2] += time_sync() - t3

        # Statistics per image
        for si, pred in enumerate(out):
            labels = targets[targets[:, 0] == si, 1:]
            nl = len(labels)
            tcls = labels[:, 0].tolist() if nl else []  # target class
            path, shape = Path(paths[si]), shapes[si][0]
            seen += 1

            if len(pred) == 0:
                if nl:
                    stats.append((torch.zeros(0, niou, dtype=torch.bool), torch.Tensor(), torch.Tensor(), tcls))
                continue

            # Predictions
            if single_cls:
                pred[:, 5] = 0
            predn = pred.clone()
            scale_coords(img[si].shape[1:], predn[:, :4], shape, shapes[si][1])  # native-space pred

            # Evaluate
            if nl:
                tbox = xywh2xyxy(labels[:, 1:5])  # target boxes
                scale_coords(img[si].shape[1:], tbox, shape, shapes[si][1])  # native-space labels
                labelsn = torch.cat((labels[:, 0:1], tbox), 1)  # native-space labels
                correct = process_batch(predn, labelsn, iouv)
                if plots:
                    confusion_matrix.process_batch(predn, labelsn)
            else:
                correct = torch.zeros(pred.shape[0], niou, dtype=torch.bool)
            stats.append((correct.cpu(), pred[:, 4].cpu(), pred[:, 5].cpu(), tcls))  # (correct, conf, pcls, tcls)

            # Save/log
            if save_txt:
                save_one_txt(predn, save_conf, shape, file=save_dir / 'labels' / (path.stem + '.txt'))
            if save_json:
                save_one_json(predn, jdict, path, class_map)  # append to COCO-JSON dictionary
            callbacks.run('on_val_image_end', pred, predn, path, names, img[si])

        # Plot images
        if plots and batch_i < 3:
            f = save_dir / f'val_batch{batch_i}_labels.jpg'  # labels
            Thread(target=plot_images, args=(img, targets, paths, f, names), daemon=True).start()
            f = save_dir / f'val_batch{batch_i}_pred.jpg'  # predictions
            Thread(target=plot_images, args=(img, output_to_target(out), paths, f, names), daemon=True).start()

    # Compute statistics
    stats = [np.concatenate(x, 0) for x in zip(*stats)]  # to numpy
    if len(stats) and stats[0].any():
        p, r, ap, f1, ap_class = ap_per_class(*stats, plot=plots, save_dir=save_dir, names=names)
        ap50, ap = ap[:, 0], ap.mean(1)  # AP@0.5, AP@0.5:0.95
        mp, mr, map50, map = p.mean(), r.mean(), ap50.mean(), ap.mean()
        nt = np.bincount(stats[3].astype(np.int64), minlength=nc)  # number of targets per class
    else:
        nt = torch.zeros(1)

    # Print results
    pf = '%20s' + '%11i' * 2 + '%11.3g' * 4  # print format
    print(pf % ('all', seen, nt.sum(), mp, mr, map50, map))

    # Print results per class
    if (verbose or (nc < 50 and not training)) and nc > 1 and len(stats):
        for i, c in enumerate(ap_class):
            print(pf % (names[c], seen, nt[c], p[i], r[i], ap50[i], ap[i]))

    # Print speeds
    t = tuple(x / seen * 1E3 for x in dt)  # speeds per image
    if not training:
        shape = (batch_size, 3, imgsz, imgsz)
        print(f'Speed: %.1fms pre-process, %.1fms inference, %.1fms NMS per image at shape {shape}' % t)

    # Plots
    if plots:
        confusion_matrix.plot(save_dir=save_dir, names=list(names.values()))
        callbacks.run('on_val_end')

    # Save JSON
    if save_json and len(jdict):
        w = Path(weights[0] if isinstance(weights, list) else weights).stem if weights is not None else ''  # weights
        anno_json = str(Path(data.get('path', '../coco')) / 'annotations/instances_val2017.json')  # annotations json
        pred_json = str(save_dir / f"{w}_predictions.json")  # predictions json
        print(f'\nEvaluating pycocotools mAP... saving {pred_json}...')
        with open(pred_json, 'w') as f:
            json.dump(jdict, f)

        try:  # https://github.com/cocodataset/cocoapi/blob/master/PythonAPI/pycocoEvalDemo.ipynb
            check_requirements(['pycocotools'])
            from pycocotools.coco import COCO
            from pycocotools.cocoeval import COCOeval

            anno = COCO(anno_json)  # init annotations api
            pred = anno.loadRes(pred_json)  # init predictions api
            eval = COCOeval(anno, pred, 'bbox')
            if is_coco:
                eval.params.imgIds = [int(Path(x).stem) for x in dataloader.dataset.img_files]  # image IDs to evaluate
            eval.evaluate()
            eval.accumulate()
            eval.summarize()
            map, map50 = eval.stats[:2]  # update results (mAP@0.5:0.95, mAP@0.5)
        except Exception as e:
            print(f'pycocotools unable to run: {e}')

    # Return results
    model.float()  # for training
    if not training:
        s = f"\n{len(list(save_dir.glob('labels/*.txt')))} labels saved to {save_dir / 'labels'}" if save_txt else ''
        print(f"Results saved to {colorstr('bold', save_dir)}{s}")
    maps = np.zeros(nc) + map
    for i, c in enumerate(ap_class):
        maps[c] = ap[i]
    return (mp, mr, map50, map, *(loss.cpu() / len(dataloader)).tolist()), maps, t


def parse_opt():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data', type=str, default=ROOT / 'data/coco128.yaml', help='dataset.yaml path')
    parser.add_argument('--weights', nargs='+', type=str, default=ROOT / 'yolov5s.pt', help='model.pt path(s)')
    parser.add_argument('--batch-size', type=int, default=32, help='batch size')
    parser.add_argument('--imgsz', '--img', '--img-size', type=int, default=640, help='inference size (pixels)')
    parser.add_argument('--conf-thres', type=float, default=0.001, help='confidence threshold')
    parser.add_argument('--iou-thres', type=float, default=0.6, help='NMS IoU threshold')
    parser.add_argument('--task', default='val', help='train, val, test, speed or study')
    parser.add_argument('--device', default='', help='cuda device, i.e. 0 or 0,1,2,3 or cpu')
    parser.add_argument('--single-cls', action='store_true', help='treat as single-class dataset')
    parser.add_argument('--augment', action='store_true', help='augmented inference')
    parser.add_argument('--verbose', action='store_true', help='report mAP by class')
    parser.add_argument('--save-txt', action='store_true', help='save results to *.txt')
    parser.add_argument('--save-hybrid', action='store_true', help='save label+prediction hybrid results to *.txt')
    parser.add_argument('--save-conf', action='store_true', help='save confidences in --save-txt labels')
    parser.add_argument('--save-json', action='store_true', help='save a COCO-JSON results file')
    parser.add_argument('--project', default=ROOT / 'runs/val', help='save to project/name')
    parser.add_argument('--name', default='exp', help='save to project/name')
    parser.add_argument('--exist-ok', action='store_true', help='existing project/name ok, do not increment')
    parser.add_argument('--half', action='store_true', help='use FP16 half-precision inference')
    opt = parser.parse_args()
    opt.data = check_yaml(opt.data)  # check YAML
    opt.save_json |= opt.data.endswith('coco.yaml')
    opt.save_txt |= opt.save_hybrid
    print_args(FILE.stem, opt)
    return opt


def main(opt):
    set_logging()
    check_requirements(exclude=('tensorboard', 'thop'))

    if opt.task in ('train', 'val', 'test'):  # run normally
        run(**vars(opt))

    elif opt.task == 'speed':  # speed benchmarks
        for w in opt.weights if isinstance(opt.weights, list) else [opt.weights]:
            run(opt.data, weights=w, batch_size=opt.batch_size, imgsz=opt.imgsz, conf_thres=.25, iou_thres=.45,
                save_json=False, plots=False)

    elif opt.task == 'study':  # run over a range of settings and save/plot
        # python val.py --task study --data coco.yaml --iou 0.7 --weights yolov5s.pt yolov5m.pt yolov5l.pt yolov5x.pt
        x = list(range(256, 1536 + 128, 128))  # x axis (image sizes)
        for w in opt.weights if isinstance(opt.weights, list) else [opt.weights]:
            f = f'study_{Path(opt.data).stem}_{Path(w).stem}.txt'  # filename to save to
            y = []  # y axis
            for i in x:  # img-size
                print(f'\nRunning {f} point {i}...')
                r, _, t = run(opt.data, weights=w, batch_size=opt.batch_size, imgsz=i, conf_thres=opt.conf_thres,
                              iou_thres=opt.iou_thres, save_json=opt.save_json, plots=False)
                y.append(r + t)  # results and times
            np.savetxt(f, y, fmt='%10.4g')  # save
        os.system('zip -r study.zip study_*.txt')
        plot_val_study(x=x)  # plot


if __name__ == "__main__":
    opt = parse_opt()
    main(opt)



"""
Module: multi_class_tracker_mixin.py

This module provides the _MultiClassTrackerMixin class, which handles tracking of 
multiple object classes using the CentroidObjectTracker.

IMPORTANT: 
	This class should be modified/tuned within individual implementations of complex ecosystems. It is recommended to
	create a `extensions.business.mixin_base.multi_class_tracker_mixin` unit that will automaticaly replace the current file

It is intended to be used as a mixin in classes that provide certain attributes and methods, including:

- self.cfg_tracking_enabled : bool
- self.cfg_tracking_mode
- self.cfg_linear_max_age
- self.cfg_linear_max_distance_ratio
- self.cfg_linear_max_relative_dist
- self.cfg_sort_min_hits
- self.cfg_sort_max_age
- self.cfg_sort_min_iou
- self.cfg_linear_max_dist_scale
- self.cfg_linear_center_dist_weight
- self.cfg_linear_hw_dist_weight
- self.cfg_linear_reset_minutes
- self._get_detector_ai_engines()
- self.dataapi_image()
- self.start_timer(name)
- self.end_timer(name)
- self.log
- self._type_to_meta_type_map

"""

import numpy as np
from datetime import datetime
from naeural_core import constants as ct
from naeural_core.utils.centroid_object_tracker import CentroidObjectTracker
from decentra_vision import geometry_methods as gmt


class _MultiClassTrackerMixin:
  """
  Mixin class for tracking multiple object classes using CentroidObjectTracker.

  This mixin depends on several attributes and methods that should be defined
  in the class that uses it. These include configuration settings, logging,
  and data API methods.

  """

  def __init__(self):
    self._object_trackers = {}
    super().__init__()

  def _match_tracking_results(self, img_inference, img_tracking_results, valid_class):
    """
    Matches all inferences to the returned objects of the tracker,
    then updates the inference dictionary with required info.

    Parameters
    ----------
    img_inference : list of dict
        List of inference dictionaries for the current image.
    img_tracking_results : dict
        Dictionary of tracking results from the tracker.
    valid_class : str
        The class of objects being tracked.

    """
    for inference in img_inference:
        inference_tlbr = np.array(self.get_inference_track_tlbr(inference)).astype(np.int32)
        for track_id, object_info in img_tracking_results.items():
            object_rectangle = np.array(object_info['rectangle']).astype(np.int32)
            if (
                np.all(inference_tlbr == object_rectangle)
                and self.__get_track_class(inference) == valid_class
            ):
                inference[ct.TRACK_ID] = track_id
                inference[ct.TRACKING_STARTED_AT] = object_info['first_update']
                alive_time = (object_info['last_update'] - object_info['first_update']).total_seconds()
                inference[ct.TRACKING_ALIVE_TIME] = alive_time
                inference[ct.APPEARANCES] = object_info['appearances']

  def __get_track_class(self, inference):
      """
      Returns the class of the object that will be tracked.
      If the object has a META_TYPE it will be used, otherwise the TYPE will be used.
      In case neither is present, None will be returned.

      Parameters
      ----------
      inference : dict
          Inference dictionary.

      Returns
      -------
      str or None
          Class of the object that will be tracked.

      """
      if self.const.META_TYPE in inference:
          return inference[self.const.META_TYPE]
      if self.const.TYPE in inference:
          return inference[self.const.TYPE]
      return None

  def get_tracking_type(self, inference):
      """
      Public method for accessing the tracking type of an inference.

      Parameters
      ----------
      inference : dict
          Inference dictionary.

      Returns
      -------
      str or None
          Tracking type of the inference.

      """
      return self.__get_track_class(inference)

  def get_inference_track_tlbr(self, inference):
      """
      Returns the TLBR that will be used for tracking an inference.
      This allows using a different TLBR for tracking than the detected one.

      Parameters
      ----------
      inference : dict
          Inference dictionary.

      Returns
      -------
      list of int
          List of 4 integers representing the TLBR used for tracking.

      """
      return inference.get(ct.TLBR_POS_TRACK, inference[ct.TLBR_POS])

  def _track_objects(self, dct_inference, img_shape=None):
      """
      Tracks inferences for multiple object classes.

      Parameters
      ----------
      dct_inference : dict
          Dictionary of inferences per AI engine.
      img_shape : tuple, optional
          Shape of the image (height, width). If None, it will be obtained from data API.

      Returns
      -------
      dict
          Updated dictionary of inferences.

      """
      if not self.cfg_tracking_enabled:
          return dct_inference

      self.start_timer('obj_track')
      detector_ai_engines = self._get_detector_ai_engines()
      inferences = []

      for engine, engine_inferences in dct_inference.items():
          if engine in detector_ai_engines:
              inferences.extend(engine_inferences)

      # Gather all valid classes
      valid_classes = set()
      for inference_list in inferences:
          for inference in inference_list:
              track_class = self.__get_track_class(inference)
              if track_class is not None:
                  valid_classes.add(track_class)

      # Ensure all previously seen classes are included
      valid_classes.update(self._object_trackers.keys())

      for valid_class in valid_classes:
          if valid_class not in self._object_trackers:
              if img_shape is None:
                  img_shape = self.dataapi_image().shape
              self._object_trackers[valid_class] = CentroidObjectTracker(
                  object_tracking_mode=self.cfg_tracking_mode,
                  linear_max_age=self.cfg_linear_max_age,
                  linear_max_distance=np.sqrt(img_shape[0] * img_shape[1]) / self.cfg_linear_max_distance_ratio,
                  linear_max_relative_distance=self.cfg_linear_max_relative_dist,
                  sort_min_hits=self.cfg_sort_min_hits,
                  sort_max_age=self.cfg_sort_max_age,
                  sort_min_iou=self.cfg_sort_min_iou,
                  max_dist_scale=self.cfg_linear_max_dist_scale,
                  center_dist_weight=self.cfg_linear_center_dist_weight,
                  hw_dist_weight=self.cfg_linear_hw_dist_weight,
                  linear_reset_minutes=self.cfg_linear_reset_minutes,
                  moved_delta_ratio=0.005  # TODO: Implement moved_delta_ratio functionality
              )

          for img_inference in inferences:
              filtered_img_inferences = [
                  inference for inference in img_inference
                  if self.__get_track_class(inference) == valid_class
              ]
              np_inferences = np.array([
                  self.get_inference_track_tlbr(inference) for inference in filtered_img_inferences
              ])
              img_tracking_results = self._object_trackers[valid_class].update_tracker(np_inferences)
              self._match_tracking_results(filtered_img_inferences, img_tracking_results, valid_class)
              self._object_trackers[valid_class].add_to_type_history(filtered_img_inferences)

      self.end_timer('obj_track')
      return dct_inference

  def _get_tracker(self, object_type):
      """
      Retrieves the tracker for a specific object type.

      Parameters
      ----------
      object_type : str
          Type or meta-type of the tracker needed.

      Returns
      -------
      CentroidObjectTracker or None
          Tracker object used for the specified type/meta-type.

      """
      if object_type in self._type_to_meta_type_map:
          object_type = self._type_to_meta_type_map[object_type]
      return self._object_trackers.get(object_type, None)

  def _get_object_appearances(self, object_id, object_type):
      """
      Returns the number of appearances for an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      int
          Number of appearances.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_object_appearances(object_id) if tracker else 0

  def _track_in_zone_objects(self, dct_inference):
      """
      Updates the time in the target zone and adds it to the inference dictionary.

      Parameters
      ----------
      dct_inference : dict
          Dictionary of inferences.

      Returns
      -------
      dict
          Updated dictionary of inferences.

      """
      if not self.cfg_tracking_enabled:
          return dct_inference

      self.start_timer('obj_track_in_zone')
      inferences = dct_inference.get(self._get_detector_ai_engine(), [])

      for obj_class, obj_tracker in self._object_trackers.items():
          for img_inference in inferences:
              filtered_img_inferences = [
                  inference for inference in img_inference
                  if self.__get_track_class(inference) == obj_class
              ]
              obj_tracker.update_in_zone_history(filtered_img_inferences)
              for inference in filtered_img_inferences:
                  inference[ct.TIME_IN_TARGET] = self.trackapi_in_zone_total_seconds(
                      object_id=inference[ct.TRACK_ID],
                      object_type=obj_class
                  )

      self.end_timer('obj_track_in_zone')
      return dct_inference

  def _get_object_history(self, object_id, object_type):
      """
      Returns the centroid history of an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      list
          List of centroids.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_object_history(object_id) if tracker else []

  def _get_object_type_history(self, object_id, object_type):
      """
      Returns the type history summary of an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      dict
          Type history summary.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_object_type_history(object_id) if tracker else {'total': 0}

  def _get_object_type_history_deque(self, object_id, object_type):
      """
      Returns the type history deque of an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      deque
          Type history deque.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_object_type_history_deque(object_id) if tracker else self.deque(maxlen=100)

  def _get_type_history_info(self, object_id, object_type):
      """
      Returns a string summarizing the type history of an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      str
          String summarizing type frequencies and recent history.

      """
      type_history = self._get_object_type_history(object_id, object_type)
      type_history_deque = self._get_object_type_history_deque(object_id, object_type)
      deque_info = [class_name[0] for class_name in type_history_deque]
      freq_info = " ".join([f"{k[:2]}:{v}" for k, v in type_history.items() if k != 'total'])
      return f'Freq: {freq_info} | Last: {"".join(deque_info)}'

  def _get_object_class_count(self, object_id, object_type, object_subtype):
      """
      Returns the count of times an object was classified as a specific subtype.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.
      object_subtype : str
          Subtype to count.

      Returns
      -------
      int
          Count of appearances as the specified subtype.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_class_count(object_id, object_subtype) if tracker else 0

  def _get_object_non_class_count(self, object_id, object_type, object_subtype):
      """
      Returns the count of times an object was not classified as a specific subtype.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.
      object_subtype : str
          Subtype to count non-appearances.

      Returns
      -------
      int
          Count of non-appearances as the specified subtype.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_non_class_count(object_id, object_subtype) if tracker else 0

  def _get_object_class_ratio(self, object_id, object_type, object_subtype):
      """
      Returns the ratio of times an object was classified as a specific subtype.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.
      object_subtype : str
          Subtype to calculate ratio for.

      Returns
      -------
      float
          Ratio of appearances as the specified subtype.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_class_ratio(object_id, object_subtype) if tracker else 0.0

  def _get_object_absolute_orientation(self, object_id, object_type):
      """
      Returns a tuple containing the original and current positions of the object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      tuple
          (original_position, current_position)

      """
      tracker = self._get_tracker(object_type)
      if not tracker:
          raise ValueError(f"No tracker found for object type '{object_type}'")
      original_position = tracker.get_original_position(object_id)
      last_position = tracker.get_object_history(object_id)[-1]
      return (original_position, last_position)

  def _get_current_orientation(self, object_id, object_type, number_of_points=3):
      """
      Returns the orientation vector based on the most recent movements.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.
      number_of_points : int, optional
          Number of recent points to consider.

      Returns
      -------
      tuple
          Orientation vector (dx, dy).

      """
      tracker = self._get_tracker(object_type)
      if not tracker:
          raise ValueError(f"No tracker found for object type '{object_type}'")
      positions = tracker.get_object_history(object_id)[-number_of_points:]
      if len(positions) < 2:
          return (0, 0)
      point1 = positions[0]
      point2 = positions[-1]
      return (point2[0] - point1[0], point2[1] - point1[1])

  def _distance_point_to_line(self, point, line):
      """
      Calculates the distance from a point to a line.

      Parameters
      ----------
      point : tuple
          Coordinates of the point (x, y).
      line : tuple
          Two points defining the line ((x1, y1), (x2, y2)).

      Returns
      -------
      float
          Distance from the point to the line.

      """
      (x0, y0) = point
      (x1, y1), (x2, y2) = line
      numerator = abs((x2 - x1) * (y1 - y0) - (x1 - x0) * (y2 - y1))
      denominator = np.hypot(x2 - x1, y2 - y1)
      return numerator / denominator if denominator != 0 else 0.0

  def _get_zone_points(self, line):
      """
      Generates two points in different zones separated by a line.

      Parameters
      ----------
      line : tuple
          Two points defining the line ((x1, y1), (x2, y2)).

      Returns
      -------
      tuple
          Two points in different zones.

      """
      (x1, y1), (x2, y2) = line
      center = ((x1 + x2) // 2, (y1 + y2) // 2)
      var_x = (x1 - x2) // 2
      var_y = (y1 - y2) // 2
      zone_1_point = (int(center[0] + var_y), int(center[1] - var_x))
      zone_2_point = (int(center[0] - var_y), int(center[1] + var_x))
      return zone_1_point, zone_2_point

  def get_movement_relative_to_line(
      self, object_id, object_type,
      line, zone1_point=None,
      zone2_point=None, threshold=10,
      start_point=None
  ):
      """
      Determines the movement of an object relative to a line.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.
      line : tuple
          Two points defining the line ((x1, y1), (x2, y2)).
      zone1_point : tuple, optional
          A point in Zone 1. If None, it will be generated automatically.
      zone2_point : tuple, optional
          A point in Zone 2. If None, it will be generated automatically.
      threshold : float, optional
          Minimum movement required to consider crossing.
      start_point : tuple, optional
          Starting point of the object. If None, the object's original position is used.

      Returns
      -------
      tuple or None
          (zone1_point, zone2_point) if moved from Zone 1 to Zone 2,
          (zone2_point, zone1_point) if moved from Zone 2 to Zone 1,
          None if movement is below threshold.

      """
      original, current = self._get_object_absolute_orientation(object_id, object_type)
      start_point = start_point or original

      distance_1 = self._distance_point_to_line(start_point, line)
      distance_2 = self._distance_point_to_line(current, line)
      distance_diff = distance_1 - distance_2

      if abs(distance_diff) < threshold:
          return None

      if zone1_point is None or zone2_point is None:
          zone1_point, zone2_point = self._get_zone_points(line)

      zone1_sign = self._distance_point_to_line(zone1_point, line) < 0
      zone2_sign = self._distance_point_to_line(zone2_point, line) < 0
      movement_sign = distance_diff < 0

      if movement_sign == zone1_sign:
          return (zone1_point, zone2_point)
      else:
          return (zone2_point, zone1_point)

  def get_line_passing_direction(
      self, object_id, object_type,
      line, zone1_point=None,
      zone2_point=None, start_point=None,
      eps=1e-5
  ):
      """
      Determines the direction in which an object passed a line.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.
      line : tuple
          Two points defining the line ((x1, y1), (x2, y2)).
      zone1_point : tuple, optional
          A point in Zone 1. If None, it will be generated automatically.
      zone2_point : tuple, optional
          A point in Zone 2. If None, it will be generated automatically.
      start_point : tuple, optional
          Starting point of the object. If None, the object's original position is used.
      eps : float, optional
          Tolerance for floating-point comparisons.

      Returns
      -------
      tuple or None
          (0, 1) if passed from Zone 1 to Zone 2,
          (1, 0) if passed from Zone 2 to Zone 1,
          None if no crossing detected.

      """
      original, current = self._get_object_absolute_orientation(object_id, object_type)
      start_point = start_point or original

      distance_1 = self._distance_point_to_line(start_point, line)
      distance_2 = self._distance_point_to_line(current, line)

      if distance_1 * distance_2 >= 0:
          return None

      movement = self.get_movement_relative_to_line(
          object_id, object_type, line, zone1_point, zone2_point, threshold=0, start_point=start_point
      )
      if movement is None:
          return None

      A, B = movement
      if np.linalg.norm(np.array(A) - np.array(zone1_point)) < eps:
          return (0, 1)
      else:
          return (1, 0)

  # TRACKAPI SECTION

  def trackapi_in_zone_total_seconds(self, object_id, object_type):
      """
      Returns the total seconds an object has been in the target zone.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      int
          Total seconds in zone.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_in_zone_total_seconds_additional(object_id) if tracker else 0

  def trackapi_in_zone_history(self, object_id, object_type):
      """
      Returns the in-zone history of an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      list
          List of intervals the object was in the target zone.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_in_zone_history(object_id) if tracker else []

  def trackapi_centroid_history(self, object_id, object_type):
      """
      Returns the centroid history of an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      list
          List of centroids.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_object_history(object_id) if tracker else []

  def trackapi_type_history(self, object_id, object_type):
      """
      Returns the type history summary of an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      dict
          Type history summary.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_object_type_history(object_id) if tracker else {'total': 0}

  def trackapi_type_history_deque(self, object_id, object_type):
      """
      Returns the type history deque of an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      deque
          Type history deque.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_object_type_history_deque(object_id) if tracker else self.deque(maxlen=100)

  def trackapi_most_seen_type(self, object_id, object_type):
      """
      Returns the most frequently observed type for an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      str
          Most seen type.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_most_seen_type(object_id) if tracker else ''

  def trackapi_class_count(self, object_id, object_type, class_name):
      """
      Returns the count of times an object was classified as a specific class.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.
      class_name : str or list of str
          Class name(s) to count.

      Returns
      -------
      int
          Count of appearances.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_class_count(object_id, class_name) if tracker else 0

  def trackapi_non_class_count(self, object_id, object_type, class_name):
      """
      Returns the count of times an object was not classified as a specific class.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.
      class_name : str or list of str
          Class name(s) to count non-appearances.

      Returns
      -------
      int
          Count of non-appearances.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_non_class_count(object_id, class_name) if tracker else 0

  def trackapi_class_ratio(self, object_id, object_type, class_name):
      """
      Returns the ratio of times an object was classified as a specific class.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.
      class_name : str or list of str
          Class name(s) to calculate ratio for.

      Returns
      -------
      float
          Ratio of appearances.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_class_ratio(object_id, class_name) if tracker else 0.0

  def trackapi_original_position(self, object_id, object_type):
      """
      Returns the original position of an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      ndarray
          Original centroid position.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_original_position(object_id) if tracker else np.array([0, 0])

  def trackapi_max_movement(self, object_id, object_type, steps=None, method='l2'):
      """
      Returns the maximum movement of an object from its original position.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.
      steps : int or None, optional
          Number of recent steps to consider.
      method : str, optional
          Distance metric to use ('l1' or 'l2').

      Returns
      -------
      float
          Maximum movement distance.

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_object_max_movement(object_id, steps, method) if tracker else 0.0

  def trackapi_last_rectangle(self, object_id, object_type):
      """
      Returns the last seen rectangle of an object.

      Parameters
      ----------
      object_id : int
          ID of the object.
      object_type : str
          Type or meta-type of the object.

      Returns
      -------
      list
          Last rectangle in the format [startX, startY, endX, endY].

      """
      tracker = self._get_tracker(object_type)
      return tracker.get_last_rectangle(object_id) if tracker else [0, 0, 0, 0]

  # END TRACKAPI SECTION


if __name__ == '__main__':
  
  eng = _MultiClassTrackerMixin()
  


"""
  This utility API for business logic plugins contains various functionalities that do not have 
  a particulat niche, providing features that otherwise would have required importing external 
  packages that are not allowed in user plugin code. 
"""
import numpy as np
from shapely import geometry
from naeural_core.business import utils

import matplotlib.pyplot as plt
import seaborn as sns

from naeural_core.main.ver import __VER__ as __CORE_VER__

try:
  from ver import __VER__ as __APP_VER__
except:
  __APP_VER__ = None

from naeural_core.utils.plugins_base.plugin_base_utils import _UtilsBaseMixin

from decentra_vision import geometry_methods


class _GenericUtilsApiMixin(_UtilsBaseMixin):

  def __init__(self):
    self.__local_data_cache = {
      'object': self.DefaultDotDict(lambda: None),
      'str': self.DefaultDotDict(lambda: ''),
      'int': self.DefaultDotDict(lambda: 0),
      'float': self.DefaultDotDict(lambda: 0.0)
    }
    super(_GenericUtilsApiMixin, self).__init__()
    return

  @property
  def geometry_methods(self):
    """Proxy for geometry_methods from decentra_vision.geometry_methods
    """
    return geometry_methods

  @property
  def gmt(self):
    """Proxy for geometry_methods from decentra_vision.geometry_methods
    """
    return self.geometry_methods

  @property
  def system_versions(self):
    return __APP_VER__, __CORE_VER__, self.log.version

  @property
  def system_version(self):
    well_defined_versions = [v for v in self.system_versions if v is not None]
    versions = '/'.join(['{}' for _ in well_defined_versions])
    versions = versions.format(*well_defined_versions)
    return "v{}".format(versions)

  @property
  def utils(self):
    """
    Provides access to methods from naeural_core.bussiness.utils.py
    """
    return utils

  @property
  def local_data_cache(self):
    """
    Can be used as a statefull store of the instance - eg `plugin.state[key]` will return `None`
    if that key has never been initialized    


    Returns
    -------
    dict
      a default dict.


    Example
    -------
      ```
      obj = self.local_data_cache['Obj1']
      if obj is None:
        obj = ClassObj1()
        self.local_data_cache['Obj1'] = obj
      ```
    """
    return self.__local_data_cache['object']

  @property
  def state(self):
    """
    Alias for `plugin.local_data_cache`
    can be used as a statefull store of the instance - eg `plugin.state[key]` will return `None`
    if that key has never been initialized     

    Returns
    -------
    dict
      Full local data cache of the current instance.

    """
    return self.local_data_cache

  @property
  def obj_cache(self):
    """
    Can be used as a statefull store of the instance - eg `plugin.obj_cache[key]` will return `None`
    if that key has never been initialized    


    Returns
    -------
    dict
      a default dict for objects.


    Example
    -------
      ```
      obj = self.obj_cache['Obj1']
      if obj is None:
        obj = ClassObj1()
        self.obj_cache['Obj1'] = obj
      ```      

    """
    return self.__local_data_cache['object']

  @property
  def int_cache(self):
    """
    can be used as a statefull store of the instance - eg `plugin.int_cache[key]` will return 0
    if that key has never been initialized    


    Returns
    -------
    dict of ints
      Returns a default dict for int values initialized with zeros.


    Example
    -------
      ```
      self.int_cache['v1'] += 1
      if self.int_cache['v1']  == 100:
        self.P("100 run interations in this plugin")
      ```

    """
    return self.__local_data_cache['int']

  @property
  def str_cache(self):
    """
    Can be used as a statefull store of the instance - eg `plugin.str_cache[key]` will return empty string
    if that key has never been initialized    


    Returns
    -------
    defaultdict
      a defaultdict with empty strings.


    Example
    -------
     ```
     self.str_cache['s1'] += str(val)[0]
     if len(self.int_cache['s1']) > 10:
       self.P("10 numbers added in the string")
     ```

    """
    return self.__local_data_cache['str']

  @property
  def float_cache(self):
    """
    Can be used as a statefull store of the instance - eg `plugin.float_cache[key]` will return 0
    if that key has never been initialized    


    Returns
    -------
    dict of floats
      Returns a default dict for float values initialized with zeros.


    Example
    -------
      ```
      self.float_cache['f1'] += val
      if self.float_cache['f1']  >= 100:
        self.P("value 100 passed")
      ```
    """
    return self.__local_data_cache['float']

  @property
  def shapely_geometry(self):
    """
    Provides access to geometry library from shapely package


    Returns
    -------
    geometry : TYPE
      DESCRIPTION.

    """
    return geometry

  @property
  def sns(self):
    """
    Provides access to the seaborn library

    Returns
    -------
    sns : package
      the Seaborn package.

    Example
    -------
      ```
      self.sns.set()
      self.sns.distplot(distribution)
      ```

    """
    return sns
  
  @property
  def pyplot(self):
    """
    Returns the matplotlib.pyplot package

    Returns
    -------
    plt : package
      the matplotlib.pyplot package.
      
    Example
    -------
      ```
      plt = self.pyplot()
      plt.plot(x, y)
      ```

    """
    return plt
  
  
  def pyplot_to_np(self, plt):
    """
    Converts a pyplot image to numpy array

    Parameters
    ----------
    plt : pyplot
      the pyplot image.

    Returns
    -------
    np.ndarray
      the numpy array image.
      
    Example
    -------
      ```
      plt = self.pyplot()
      plt.plot(x, y)
      img = self.pyplot_to_np(plt)
      ```

    """
    return self.log.plt_to_np(plt)

  def should_progress(self, progress, step=5):
    """
    Helper function for progress intervals from 5 to 5%. Returns true if param progress hits the value
    else false. Once a `True` is returned it will never again be returned

    Parameters
    ----------
    progress : float
      percentage 0-100.

    Returns
    -------
    result : bool
      a milestone is reached or not.

    """
    if not hasattr(self, '_dct_progress_intervals'):
      self._dct_progress_intervals = {i: True for i in list(range(1, 101, step))}
      self.P("Created progress intervals at step {}".format(step))
    progress = int(progress)
    progress_status = self._dct_progress_intervals.get(progress, False)
    result = False
    if isinstance(progress_status, bool) and progress_status:
      self._dct_progress_intervals[progress] = self.time()
      result = True
    return result

  def get_alive_time(self, as_str=False):
    """
    Returns plugin alive time

    Parameters
    ----------
    as_str : bool, optional
      return as string. The default is False.

    Returns
    -------
    result : float or str


    Example
    -------
      ```
      result = 'Plugin was alive ' + self.get_alive_time(as_str=True)
      ```
    """
    res = self.time_alive
    if as_str:
      res = str(self.timedelta(seconds=int(self.time_alive)))
    return res

  def get_exception(self):
    """
    Returns last exception fullstack

    Returns
    -------
    string
      The full multi-line stack.

    Example:
      ```
      ```

    """
    import traceback
    return traceback.format_exc()

  def get_models_file(self, fn):
    """
    Retruns path to models file

    :param fn: string - file name

    """
    return self.log.get_models_file(fn)

  def create_basic_ts_model(self, series_min: int = 100, train_hist=None, train_periods=None):
    """
    Returns a basic time-series prediction model instance

    Parameters
    ----------
    series_min : int, optional
      Minimal accepted number of historical steps. The default is 100.
    train_hist : int, optional
      The training window size. The default is None.
    train_periods : int, optional
      how many windows to use. The default is None.

    Returns
    -------
    BasicSeriesModel() object


    Example
    -------
      ```
        # init model
        model = plugin.create_basic_ts_model(...)
      ```

    """
    from naeural_core.utils.basic_series_model import BasicSeriesModel

    return BasicSeriesModel(
      series_min=series_min,
      train_hist=train_hist,
      train_periods=train_periods
    )

  def basic_ts_create(self, series_min=100, train_hist=None, train_periods=None):
    """
    Returns a basic time-series prediction model instance

    Parameters
    ----------
    series_min : int, optional
      Minimal accepted number of historical steps. The default is 100.
    train_hist : int, optional
      The training window size. The default is None.
    train_periods : int, optional
      how many windows to use. The default is None.

    Returns
    -------
    BasicSeriesModel() object


    Example
    -------
      ```
        # init model
        model = plugin.basic_ts_create(...)
      ```

    """
    return self.create_basic_ts_model(
      series_min=series_min,
      train_hist=train_hist,
      train_periods=train_periods
    )

  def basic_ts_fit_predict(self, data, steps):
    """
    Takes a list of values and directly returns predictions using a basic AR model


    Parameters
    ----------
    data : list
      list of float values.
    steps : int
      number of prediction steps.

    Returns
    -------
    yh : list
      the `steps` predicted values.


    Example
    -------
      ```
      yh = self.basic_ts_fit_predict(data=[0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89], steps=3)
      result = {'preds' : yh}
      ```

    """

    model = self.create_basic_ts_model(series_min=len(data))
    model.fit(data)
    yh = model.predict(steps)
    return yh
  
  ## MLAPI
  if True:
    def mlapi_timeseries_fit_predict(self, data, steps, **kwargs):
      """
      Takes a list of values and directly returns predictions using a basic AR model


      Parameters
      ----------
      data : list
        list of float values.
      steps : int
        number of prediction steps.

      Returns
      -------
      yh : list
        the `steps` predicted values.


      Example
      -------
        ```
        yh = self.basic_ts_fit_predict(data=[0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89], steps=3)
        result = {'preds' : yh}
        ```

      """
      result = self.basic_ts_fit_predict(data, steps)
      return result
    
    
    def mlapi_create_ts_model(self, series_min=100, train_hist=None, train_periods=None):
      """
      Returns a basic time-series prediction model instance

      Parameters
      ----------
      series_min : int, optional
        Minimal accepted number of historical steps. The default is 100.
      train_hist : int, optional
        The training window size. The default is None.
      train_periods : int, optional
        how many windows to use. The default is None.

      Returns
      -------
      BasicSeriesModel() object


      Example
      -------
        ```
          # init model
          model = plugin.basic_ts_create(...)
        ```

      """
      return self.create_basic_ts_model(
        series_min=series_min,
        train_hist=train_hist,
        train_periods=train_periods
      )
      
    def mlapi_create_anomaly_model(self):
      """
      Returns a basic anomaly model instance

      Returns
      -------
      BasicAnomalyModel() object


      Example
      -------
        ```
          # init model
          model = plugin.mlapi_create_anomaly_model(...)
        ```

      """
      from naeural_core.utils.basic_anomaly_model import BasicAnomalyModel
      return BasicAnomalyModel()
    
    
    def mlapi_anomaly_fit_predict(self, x_train, x_test, proba=True):
      """
      Takes a list of values and directly returns predictions using a basic anomaly detection model


      Parameters
      ----------
      x_train : list
        list of float values. These are the training values.
        
      x_test : list
        list of float values. These are the test values.
        
      proba : bool, optional
        If `True` then the model will return the probability of being an anomaly. The default is True.

      Returns
      -------
      yh : list
        the `steps` predicted values.


      Example
      -------
        ```
        yproba = self.mlapi_anomaly_fit_predict(x_train=[0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89], x_test=[0, 1, 1, 2, 3], proba=True)
        result = {'preds' : yproba}
        ```

      """
      model = self.mlapi_create_anomaly_model()
      model.fit(x_train)
      result = model.predict(x_test, proba=proba)
      return result
    
    
  ## END MLAPI

  def create_sre(self, **kwargs):
    """
    Returns a Statefull Rule Engine instance


    Returns
    -------
    SRE()


    Example
    -------
      ```
      eng = self.create_sre()  
      # add a data stream
      eng.add_entity(
        entity_id='dev_test_1', 
        entity_props=['f1','f2','f3'],
        entity_rules=['state.f1.val == 0 and state.f2.val == 0 and prev.f2.val==1'],
      )

      ```

    """
    # TODO: encapsulate this in try-except + change location to extensions
    try:
      from extensions.utils.sre import SRE
      return SRE(log=self.log, **kwargs)
    except:
      return None

  def create_statefull_rule_engine(self, **kwargs):
    """
    Returns a Statefull Rule Engine instance


    Returns
    -------
    SRE()


    Example
    -------
      ```
      eng = self.create_statefull_rule_engine()  
      # add a data stream
      eng.add_entity(
        entity_id='dev_test_1', 
        entity_props=['f1','f2','f3'],
        entity_rules=['state.f1.val == 0 and state.f2.val == 0 and prev.f2.val==1'],
      )

      ```

    """
    return self.create_sre(**kwargs)

  def plot_ts(self, vals, vals_pred=None, title=''):
    """
    Generates a `default_image` that will be embedded in the plugin response containing
    a time-series plot


    Parameters
    ----------
    vals : list[float]
      the backlog data.
    vals_pred : list[float], optional
      prediction data. The default is None.
    title : str, optional
      a title for our plot. The default is ''.

    Returns
    -------
    msg : str
      a error or success `'Plot ok'` message.


    Example
    -------
      ```
      vals = [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89]
      steps = 3
      yh = self.basic_ts_fit_predict(data=vals, steps=steps)
      has_plot = self.plot_ts(vals, vals_pred=yh, title='prediction for {} steps'.format(steps))
      result = {'preds' : yh, 'plot_ok' : has_plot}
      ```


    """
    self.maybe_start_thread_safe_drawing()
    try:
      import matplotlib.pyplot as plt
      import seaborn as sns
      sns.set()
      sns.set(font_scale=2)
      plt.figure(figsize=(30, 14))
      l1 = len(vals)
      min_tick = 1
      max_tick = l1
      plt.plot(np.arange(min_tick, max_tick + 1), vals, 'o--', c='b', label='data')
      if vals_pred is not None and len(vals_pred) > 0:
        l2 = len(vals_pred)
        max_tick = l1 + l2
        plt.plot(np.arange(l1, max_tick + 1), [vals[-1]] + vals_pred, 'o--', c='r', label='prediction')
      plt.xticks(np.arange(min_tick, max_tick + 1))
      plt.xlabel('Steps')
      plt.title(title)
      plt.legend()
      img = self.log.plt_to_np(plt, axis=True)
      self.set_output_image(img)
      self.log.unlock_resource('api_safe_plot')
      msg = 'Plot ok.'
    except Exception as e:
      msg = str(e)
    self.maybe_stop_thread_safe_drawing()
    return msg

  def vision_plot_detections(self):
    """
    Plots detection on default output image if any


    Returns
    -------
    None.


    Example
    -------
      ```
      img = self.dataapi_image()
      if img is not None: # no need to try and plot if there is not image        
        self.vision_plot_detections()
      ```

    """
    instance_inferences = self.dataapi_image_instance_inferences()
    if instance_inferences is not None and len(instance_inferences) > 0:
      img = self.dataapi_image()
      img = self._painter.draw_inference_boxes(
        image=img.copy(),
        lst_inf=instance_inferences
      )
      self.set_output_image(img)
    return

  def set_default_image(self, img=None):
    """
    Sets given image as witness for current payload

    Parameters
    ----------
    img : np.ndarray
      the RGB image.


    Example
    -------
      ```
      img = self.dataapi_image()
      self.set_default_image(img)
      ```

    """
    if img is None:
      img = self.dataapi_image()
    self.set_output_image(img)
    return

  def set_witness_image(self, img=None):
    """
    Sets given image as witness for current payload

    Parameters
    ----------
    img : np.ndarray
      the RGB image.


    Example
    -------
      ```
      img = self.dataapi_image()
      self.set_witness_image(img)
      ```

    """
    self.set_default_image(img)
    return

  def set_text_witness(self, text: str):
    """
    Creates a simple empty witness with given centered text.

    Parameters
    ----------
    text : str
      The text that will be in the output. If the text is bigger than the screen 
      it will be displayed on multiple lines

    Returns
    -------
    None.


    Example
    -------
      ```
      self.set_text_witness('Hello world in witness image :)')
      ```


    """
    H, W = 720, 1280
    left = 100
    text = self.normalize_text(text)
    font = self.cv2.FONT_HERSHEY_SIMPLEX
    size = 0.7
    thickness = 2
    font_scale = size
    tw, _ = self._painter.text_size(text, font, font_scale, thickness)[0]
    if tw > (W - left):
      segs = tw // (W - left * 2) + 1
      words = text.split()
      n_words = len(words)
      nwps = n_words // segs
      remain = n_words % segs
      parts = []
      sidx = 0
      for i in range(segs):
        eidx = sidx + nwps
        parts.append(' '.join(words[sidx:eidx]))
        sidx = eidx
      if remain > 0:
        parts.append(' '.join(words[sidx:sidx + remain]))
      text = parts
    img = np.zeros((H, W, 3), dtype='uint8')
    img = self._painter.alpha_text_rectangle(
      image=img,
      text=text,
      left=left,
      top=300,
      color=(0, 250, 0),
      font=font,
      size=size,
      thickness=thickness,
    )
    self.set_default_image(img=img)
    return

  def chatapi_ask(self, question: str, persona: str, user: str, set_witness: bool = True, personas_folder: str = None):
    """
    Simple single-function API for accessing chat backend. Provides statefullness based on
    provided `user` for the caller plugin instance.

    Parameters
    ----------
    question : str
      The question.
    persona : str
      A valid persona.
    user : str
      A user name for tracking your session.
    set_witness : bool, optional
      If `True` then a witness will be generated. The default is True.

    Returns
    -------
    result : str
      The response.


    Example
    -------
      ```
      result = plugin.chatapi_ask(
        question="Who are you?",
        persona='codegen',
        user="John Doe",
      )      
      ```

    """
    from naeural_core.utils.openai.app import OpenAIApp, PERSONAS_FOLDER
    result = None
    assert isinstance(question, str) and len(question) > 1, "Please ask a valid question"
    assert isinstance(persona, str) and len(persona) > 1, "Please give a valid persona"
    assert isinstance(user, str) and len(user) > 1, "Please give a valid user name"
    personas_folder = personas_folder or PERSONAS_FOLDER
    eng_id = 'chat_eng_' + user
    eng = self.obj_cache[eng_id]
    if eng is None:
      eng = OpenAIApp(persona=persona, user=user, persona_location=personas_folder)
      self.obj_cache[eng_id] = eng
    self.P("Executing `{}.ask()` for user '{}' with persona '{}'...".format(eng_id, user, persona))
    t0 = self.time()
    result = eng.ask(question)
    elapsed = self.time() - t0
    self.P("  Executed `{}.ask()` in {:.2f}s seconds".format(eng_id, elapsed))
    if set_witness and self._painter is not None:
      self.set_text_witness(text=result)
    return result

  def get_serving_processes(self):
    """
    Returns a list of used AI Engines within the current plugin instance based on given configuration

    Parameters
    ----------
    None.

    Returns
    -------
    result : list
      The list.


    Example
    -------
      ```
      lst_servers = plugin.get_serving_processes()
      ```
    """
    engines = self.cfg_ai_engine
    if not isinstance(engines, list):
      engines = [engines]
    return [self.get_serving_process_given_ai_engine(x) for x in engines]

  # payload section

  def payload_set_value(self, key: str, val):
    """
    This method allows the addition of data directly in the next outgoing payload
    from the current biz plugin instance

    Parameters
    ----------
    key : str
      the name of the key
    val : any
      A value that will be json-ified.

    Returns
    -------
    None.


    Example:
    -------      
      ```
      bool_is_alert = ...
      plugin.payload_set_value("is_special_alert", bool_is_alert)
      ```

    """
    self.payload_set_data(key, val)
    return

  def download(self, url, fn, target='output', **kwargs):
    """
    Dowload wrapper that will download a given file from a url to `_local_cache/_output.


    TODO: fix to use specific endpoints configs not only from base file_system_manager

    Parameters
    ----------
    url : str
      the url where to find the file.

    fn : str
      local file name to be saved in `target` folder.

    **kwargs : dict
      params for special upload procedures such as minio.


    Returns
    -------
    res : str
      path of the downloaded file, None if not found.


    Example
    -------

      ```
      res = plugin.download('http://drive.google.com/file-url', 'test.bin')
      if res is not None:
        plugin.P("Downloaded!")
      ```

    """
    # this bypasses file-manager and only uses the config data due to incomplete feats in file-mgr
    file_system = self.global_shmem['file_system_manager']._file_system
    fs_args = {k.lower(): v for k, v in file_system.config_data.items()}
    all_args = {
      **fs_args,
      **kwargs,
    }
    self.P("Running download '{}' with arguments '{}' based on {}...".format(url, all_args, file_system.__class__.__name__))
    res = self.maybe_download(
      url=url,
      fn=fn,
      target=target,
      **all_args,
    )
    return res

  # end payload section



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Loss functions
"""

import torch
import torch.nn as nn

from utils.metrics import bbox_iou
from utils.torch_utils import is_parallel


def smooth_BCE(eps=0.1):  # https://github.com/ultralytics/yolov3/issues/238#issuecomment-598028441
    # return positive, negative label smoothing BCE targets
    return 1.0 - 0.5 * eps, 0.5 * eps


class BCEBlurWithLogitsLoss(nn.Module):
    # BCEwithLogitLoss() with reduced missing label effects.
    def __init__(self, alpha=0.05):
        super(BCEBlurWithLogitsLoss, self).__init__()
        self.loss_fcn = nn.BCEWithLogitsLoss(reduction='none')  # must be nn.BCEWithLogitsLoss()
        self.alpha = alpha

    def forward(self, pred, true):
        loss = self.loss_fcn(pred, true)
        pred = torch.sigmoid(pred)  # prob from logits
        dx = pred - true  # reduce only missing label effects
        # dx = (pred - true).abs()  # reduce missing label and false label effects
        alpha_factor = 1 - torch.exp((dx - 1) / (self.alpha + 1e-4))
        loss *= alpha_factor
        return loss.mean()


class FocalLoss(nn.Module):
    # Wraps focal loss around existing loss_fcn(), i.e. criteria = FocalLoss(nn.BCEWithLogitsLoss(), gamma=1.5)
    def __init__(self, loss_fcn, gamma=1.5, alpha=0.25):
        super(FocalLoss, self).__init__()
        self.loss_fcn = loss_fcn  # must be nn.BCEWithLogitsLoss()
        self.gamma = gamma
        self.alpha = alpha
        self.reduction = loss_fcn.reduction
        self.loss_fcn.reduction = 'none'  # required to apply FL to each element

    def forward(self, pred, true):
        loss = self.loss_fcn(pred, true)
        # p_t = torch.exp(-loss)
        # loss *= self.alpha * (1.000001 - p_t) ** self.gamma  # non-zero power for gradient stability

        # TF implementation https://github.com/tensorflow/addons/blob/v0.7.1/tensorflow_addons/losses/focal_loss.py
        pred_prob = torch.sigmoid(pred)  # prob from logits
        p_t = true * pred_prob + (1 - true) * (1 - pred_prob)
        alpha_factor = true * self.alpha + (1 - true) * (1 - self.alpha)
        modulating_factor = (1.0 - p_t) ** self.gamma
        loss *= alpha_factor * modulating_factor

        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        else:  # 'none'
            return loss


class QFocalLoss(nn.Module):
    # Wraps Quality focal loss around existing loss_fcn(), i.e. criteria = FocalLoss(nn.BCEWithLogitsLoss(), gamma=1.5)
    def __init__(self, loss_fcn, gamma=1.5, alpha=0.25):
        super(QFocalLoss, self).__init__()
        self.loss_fcn = loss_fcn  # must be nn.BCEWithLogitsLoss()
        self.gamma = gamma
        self.alpha = alpha
        self.reduction = loss_fcn.reduction
        self.loss_fcn.reduction = 'none'  # required to apply FL to each element

    def forward(self, pred, true):
        loss = self.loss_fcn(pred, true)

        pred_prob = torch.sigmoid(pred)  # prob from logits
        alpha_factor = true * self.alpha + (1 - true) * (1 - self.alpha)
        modulating_factor = torch.abs(true - pred_prob) ** self.gamma
        loss *= alpha_factor * modulating_factor

        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        else:  # 'none'
            return loss


class ComputeLoss:
    # Compute losses
    def __init__(self, model, autobalance=False):
        self.sort_obj_iou = False
        device = next(model.parameters()).device  # get model device
        h = model.hyp  # hyperparameters

        # Define criteria
        BCEcls = nn.BCEWithLogitsLoss(pos_weight=torch.tensor([h['cls_pw']], device=device))
        BCEobj = nn.BCEWithLogitsLoss(pos_weight=torch.tensor([h['obj_pw']], device=device))

        # Class label smoothing https://arxiv.org/pdf/1902.04103.pdf eqn 3
        self.cp, self.cn = smooth_BCE(eps=h.get('label_smoothing', 0.0))  # positive, negative BCE targets

        # Focal loss
        g = h['fl_gamma']  # focal loss gamma
        if g > 0:
            BCEcls, BCEobj = FocalLoss(BCEcls, g), FocalLoss(BCEobj, g)

        det = model.module.model[-1] if is_parallel(model) else model.model[-1]  # Detect() module
        self.balance = {3: [4.0, 1.0, 0.4]}.get(det.nl, [4.0, 1.0, 0.25, 0.06, .02])  # P3-P7
        self.ssi = list(det.stride).index(16) if autobalance else 0  # stride 16 index
        self.BCEcls, self.BCEobj, self.gr, self.hyp, self.autobalance = BCEcls, BCEobj, 1.0, h, autobalance
        for k in 'na', 'nc', 'nl', 'anchors':
            setattr(self, k, getattr(det, k))

    def __call__(self, p, targets):  # predictions, targets, model
        device = targets.device
        lcls, lbox, lobj = torch.zeros(1, device=device), torch.zeros(1, device=device), torch.zeros(1, device=device)
        tcls, tbox, indices, anchors = self.build_targets(p, targets)  # targets

        # Losses
        for i, pi in enumerate(p):  # layer index, layer predictions
            b, a, gj, gi = indices[i]  # image, anchor, gridy, gridx
            tobj = torch.zeros_like(pi[..., 0], device=device)  # target obj

            n = b.shape[0]  # number of targets
            if n:
                ps = pi[b, a, gj, gi]  # prediction subset corresponding to targets

                # Regression
                pxy = ps[:, :2].sigmoid() * 2. - 0.5
                pwh = (ps[:, 2:4].sigmoid() * 2) ** 2 * anchors[i]
                pbox = torch.cat((pxy, pwh), 1)  # predicted box
                iou = bbox_iou(pbox.T, tbox[i], x1y1x2y2=False, CIoU=True)  # iou(prediction, target)
                lbox += (1.0 - iou).mean()  # iou loss

                # Objectness
                score_iou = iou.detach().clamp(0).type(tobj.dtype)
                if self.sort_obj_iou:
                    sort_id = torch.argsort(score_iou)
                    b, a, gj, gi, score_iou = b[sort_id], a[sort_id], gj[sort_id], gi[sort_id], score_iou[sort_id]
                tobj[b, a, gj, gi] = (1.0 - self.gr) + self.gr * score_iou  # iou ratio

                # Classification
                if self.nc > 1:  # cls loss (only if multiple classes)
                    t = torch.full_like(ps[:, 5:], self.cn, device=device)  # targets
                    t[range(n), tcls[i]] = self.cp
                    lcls += self.BCEcls(ps[:, 5:], t)  # BCE

                # Append targets to text file
                # with open('targets.txt', 'a') as file:
                #     [file.write('%11.5g ' * 4 % tuple(x) + '\n') for x in torch.cat((txy[i], twh[i]), 1)]

            obji = self.BCEobj(pi[..., 4], tobj)
            lobj += obji * self.balance[i]  # obj loss
            if self.autobalance:
                self.balance[i] = self.balance[i] * 0.9999 + 0.0001 / obji.detach().item()

        if self.autobalance:
            self.balance = [x / self.balance[self.ssi] for x in self.balance]
        lbox *= self.hyp['box']
        lobj *= self.hyp['obj']
        lcls *= self.hyp['cls']
        bs = tobj.shape[0]  # batch size

        return (lbox + lobj + lcls) * bs, torch.cat((lbox, lobj, lcls)).detach()

    def build_targets(self, p, targets):
        # Build targets for compute_loss(), input targets(image,class,x,y,w,h)
        na, nt = self.na, targets.shape[0]  # number of anchors, targets
        tcls, tbox, indices, anch = [], [], [], []
        gain = torch.ones(7, device=targets.device)  # normalized to gridspace gain
        ai = torch.arange(na, device=targets.device).float().view(na, 1).repeat(1, nt)  # same as .repeat_interleave(nt)
        targets = torch.cat((targets.repeat(na, 1, 1), ai[:, :, None]), 2)  # append anchor indices

        g = 0.5  # bias
        off = torch.tensor([[0, 0],
                            [1, 0], [0, 1], [-1, 0], [0, -1],  # j,k,l,m
                            # [1, 1], [1, -1], [-1, 1], [-1, -1],  # jk,jm,lk,lm
                            ], device=targets.device).float() * g  # offsets

        for i in range(self.nl):
            anchors = self.anchors[i]
            gain[2:6] = torch.tensor(p[i].shape)[[3, 2, 3, 2]]  # xyxy gain

            # Match targets to anchors
            t = targets * gain
            if nt:
                # Matches
                r = t[:, :, 4:6] / anchors[:, None]  # wh ratio
                j = torch.max(r, 1. / r).max(2)[0] < self.hyp['anchor_t']  # compare
                # j = wh_iou(anchors, t[:, 4:6]) > model.hyp['iou_t']  # iou(3,n)=wh_iou(anchors(3,2), gwh(n,2))
                t = t[j]  # filter

                # Offsets
                gxy = t[:, 2:4]  # grid xy
                gxi = gain[[2, 3]] - gxy  # inverse
                j, k = ((gxy % 1. < g) & (gxy > 1.)).T
                l, m = ((gxi % 1. < g) & (gxi > 1.)).T
                j = torch.stack((torch.ones_like(j), j, k, l, m))
                t = t.repeat((5, 1, 1))[j]
                offsets = (torch.zeros_like(gxy)[None] + off[:, None])[j]
            else:
                t = targets[0]
                offsets = 0

            # Define
            b, c = t[:, :2].long().T  # image, class
            gxy = t[:, 2:4]  # grid xy
            gwh = t[:, 4:6]  # grid wh
            gij = (gxy - offsets).long()
            gi, gj = gij.T  # grid xy indices

            # Append
            a = t[:, 6].long()  # anchor indices
            indices.append((b, a, gj.clamp_(0, gain[3] - 1), gi.clamp_(0, gain[2] - 1)))  # image, anchor, grid indices
            tbox.append(torch.cat((gxy - gij, gwh), 1))  # box
            anch.append(anchors[a])  # anchors
            tcls.append(c)  # class

        return tcls, tbox, indices, anch



import numpy as np


def get_model_size(model):
  n_params = 0
  n_bytes = 0
  for param in model.parameters():
    npar = param.nelement()
    size = npar * param.element_size()
    n_bytes += size
    n_params += npar
  return n_params, n_bytes


def get_dropout_rate(
    value_type='constant',
    dropout=0.1,
    step=0,
    max_step=0
):
  """
  Utility method for getting the dropout rate for a multistep training process.
  Parameters
  ----------
  value_type : str,
    if 'constant' the dropout rate will be constant,
    if 'last' the dropout rate will be 0 until the last step,
    if 'progressive' the dropout rate will increase linearly from 0 to the specified dropout rate.
    otherwise the dropout rate will be 0
  dropout : float, the base dropout rate
  step : int, the current step in the training process
  max_step : int, the maximum number of steps in the training process

  Returns
  -------
  float, the computed dropout rate
  """
  dropout = max(0.0, min(1.0, dropout))
  if value_type == 'constant':
    return dropout
  if value_type == 'last':
    return 0 if step < max_step else dropout
  if value_type == 'progressive':
    return np.linspace(dropout / 10, dropout, max_step + 1)[step]
  return 0


def conv_output_shape(
    h_w, 
    kernel_size=1, 
    stride=1, 
    pad=0, 
    dilation=1, 
    transpose=False,
    **kwargs
  ):
  """
  Utility function for computing output of convolutions
  takes a tuple of (h,w) and returns a tuple of (h,w)

  credits: https://discuss.pytorch.org/t/utility-function-for-calculating-the-shape-of-a-conv-output/11173/6
  """

  if isinstance(h_w, (tuple, list)):
    if len(h_w) == 3:
      h_w = h_w[1:]
  elif isinstance(h_w, int):
    h_w = (h_w, h_w)
  
  if isinstance(kernel_size, int):
    kernel_size = (kernel_size, kernel_size)

  if isinstance(stride, int):
    stride = (stride, stride)

  if 'padding' in kwargs:
    pad = kwargs['padding']
    
  if isinstance(pad, int):
    pad = (pad, pad)

  if isinstance(dilation, int):
    dilation = (dilation, dilation)
    
  h, w = h_w
  kh, kw = kernel_size
  sh, sw = stride
  ph, pw = pad
  dh, dw = dilation
  if not transpose:
    h = (h + 2 * ph - dh * (kh - 1) - 1) // sh + 1
    w = (w + 2 * pw - dw * (kw - 1) - 1) // sw + 1
  else:
    h = (h - 1) * sh - 2 * ph + (kh - 1) + 1    
    w = (w - 1) * sw - 2 * pw + (kw - 1) + 1    
  return h, w


# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
TensorFlow, Keras and TFLite versions of YOLOv5
Authored by https://github.com/zldrobit in PR https://github.com/ultralytics/yolov5/pull/1127

Usage:
    $ python models/tf.py --weights yolov5s.pt

Export:
    $ python path/to/export.py --weights yolov5s.pt --include saved_model pb tflite tfjs
"""

import argparse
import logging
import sys
from copy import deepcopy
from pathlib import Path

FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]  # YOLOv5 root directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH
# ROOT = ROOT.relative_to(Path.cwd())  # relative

import numpy as np
import tensorflow as tf
import torch
import torch.nn as nn
from tensorflow import keras

from models.common import Conv, Bottleneck, SPP, DWConv, Focus, BottleneckCSP, Concat, autopad, C3
from models.experimental import CrossConv, MixConv2d, attempt_load
from models.yolo import Detect
from utils.general import make_divisible, print_args, set_logging
from utils.activations import SiLU

LOGGER = logging.getLogger(__name__)


class TFBN(keras.layers.Layer):
    # TensorFlow BatchNormalization wrapper
    def __init__(self, w=None):
        super(TFBN, self).__init__()
        self.bn = keras.layers.BatchNormalization(
            beta_initializer=keras.initializers.Constant(w.bias.numpy()),
            gamma_initializer=keras.initializers.Constant(w.weight.numpy()),
            moving_mean_initializer=keras.initializers.Constant(w.running_mean.numpy()),
            moving_variance_initializer=keras.initializers.Constant(w.running_var.numpy()),
            epsilon=w.eps)

    def call(self, inputs):
        return self.bn(inputs)


class TFPad(keras.layers.Layer):
    def __init__(self, pad):
        super(TFPad, self).__init__()
        self.pad = tf.constant([[0, 0], [pad, pad], [pad, pad], [0, 0]])

    def call(self, inputs):
        return tf.pad(inputs, self.pad, mode='constant', constant_values=0)


class TFConv(keras.layers.Layer):
    # Standard convolution
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, act=True, w=None):
        # ch_in, ch_out, weights, kernel, stride, padding, groups
        super(TFConv, self).__init__()
        assert g == 1, "TF v2.2 Conv2D does not support 'groups' argument"
        assert isinstance(k, int), "Convolution with multiple kernels are not allowed."
        # TensorFlow convolution padding is inconsistent with PyTorch (e.g. k=3 s=2 'SAME' padding)
        # see https://stackoverflow.com/questions/52975843/comparing-conv2d-with-padding-between-tensorflow-and-pytorch

        conv = keras.layers.Conv2D(
            c2, k, s, 'SAME' if s == 1 else 'VALID', use_bias=False if hasattr(w, 'bn') else True,
            kernel_initializer=keras.initializers.Constant(w.conv.weight.permute(2, 3, 1, 0).numpy()),
            bias_initializer='zeros' if hasattr(w, 'bn') else keras.initializers.Constant(w.conv.bias.numpy()))
        self.conv = conv if s == 1 else keras.Sequential([TFPad(autopad(k, p)), conv])
        self.bn = TFBN(w.bn) if hasattr(w, 'bn') else tf.identity

        # YOLOv5 activations
        if isinstance(w.act, nn.LeakyReLU):
            self.act = (lambda x: keras.activations.relu(x, alpha=0.1)) if act else tf.identity
        elif isinstance(w.act, nn.Hardswish):
            self.act = (lambda x: x * tf.nn.relu6(x + 3) * 0.166666667) if act else tf.identity
        elif isinstance(w.act, (nn.SiLU, SiLU)):
            self.act = (lambda x: keras.activations.swish(x)) if act else tf.identity
        else:
            raise Exception(f'no matching TensorFlow activation found for {w.act}')

    def call(self, inputs):
        return self.act(self.bn(self.conv(inputs)))


class TFFocus(keras.layers.Layer):
    # Focus wh information into c-space
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, act=True, w=None):
        # ch_in, ch_out, kernel, stride, padding, groups
        super(TFFocus, self).__init__()
        self.conv = TFConv(c1 * 4, c2, k, s, p, g, act, w.conv)

    def call(self, inputs):  # x(b,w,h,c) -> y(b,w/2,h/2,4c)
        # inputs = inputs / 255.  # normalize 0-255 to 0-1
        return self.conv(tf.concat([inputs[:, ::2, ::2, :],
                                    inputs[:, 1::2, ::2, :],
                                    inputs[:, ::2, 1::2, :],
                                    inputs[:, 1::2, 1::2, :]], 3))


class TFBottleneck(keras.layers.Layer):
    # Standard bottleneck
    def __init__(self, c1, c2, shortcut=True, g=1, e=0.5, w=None):  # ch_in, ch_out, shortcut, groups, expansion
        super(TFBottleneck, self).__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = TFConv(c1, c_, 1, 1, w=w.cv1)
        self.cv2 = TFConv(c_, c2, 3, 1, g=g, w=w.cv2)
        self.add = shortcut and c1 == c2

    def call(self, inputs):
        return inputs + self.cv2(self.cv1(inputs)) if self.add else self.cv2(self.cv1(inputs))


class TFConv2d(keras.layers.Layer):
    # Substitution for PyTorch nn.Conv2D
    def __init__(self, c1, c2, k, s=1, g=1, bias=True, w=None):
        super(TFConv2d, self).__init__()
        assert g == 1, "TF v2.2 Conv2D does not support 'groups' argument"
        self.conv = keras.layers.Conv2D(
            c2, k, s, 'VALID', use_bias=bias,
            kernel_initializer=keras.initializers.Constant(w.weight.permute(2, 3, 1, 0).numpy()),
            bias_initializer=keras.initializers.Constant(w.bias.numpy()) if bias else None, )

    def call(self, inputs):
        return self.conv(inputs)


class TFBottleneckCSP(keras.layers.Layer):
    # CSP Bottleneck https://github.com/WongKinYiu/CrossStagePartialNetworks
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5, w=None):
        # ch_in, ch_out, number, shortcut, groups, expansion
        super(TFBottleneckCSP, self).__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = TFConv(c1, c_, 1, 1, w=w.cv1)
        self.cv2 = TFConv2d(c1, c_, 1, 1, bias=False, w=w.cv2)
        self.cv3 = TFConv2d(c_, c_, 1, 1, bias=False, w=w.cv3)
        self.cv4 = TFConv(2 * c_, c2, 1, 1, w=w.cv4)
        self.bn = TFBN(w.bn)
        self.act = lambda x: keras.activations.relu(x, alpha=0.1)
        self.m = keras.Sequential([TFBottleneck(c_, c_, shortcut, g, e=1.0, w=w.m[j]) for j in range(n)])

    def call(self, inputs):
        y1 = self.cv3(self.m(self.cv1(inputs)))
        y2 = self.cv2(inputs)
        return self.cv4(self.act(self.bn(tf.concat((y1, y2), axis=3))))


class TFC3(keras.layers.Layer):
    # CSP Bottleneck with 3 convolutions
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5, w=None):
        # ch_in, ch_out, number, shortcut, groups, expansion
        super(TFC3, self).__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = TFConv(c1, c_, 1, 1, w=w.cv1)
        self.cv2 = TFConv(c1, c_, 1, 1, w=w.cv2)
        self.cv3 = TFConv(2 * c_, c2, 1, 1, w=w.cv3)
        self.m = keras.Sequential([TFBottleneck(c_, c_, shortcut, g, e=1.0, w=w.m[j]) for j in range(n)])

    def call(self, inputs):
        return self.cv3(tf.concat((self.m(self.cv1(inputs)), self.cv2(inputs)), axis=3))


class TFSPP(keras.layers.Layer):
    # Spatial pyramid pooling layer used in YOLOv3-SPP
    def __init__(self, c1, c2, k=(5, 9, 13), w=None):
        super(TFSPP, self).__init__()
        c_ = c1 // 2  # hidden channels
        self.cv1 = TFConv(c1, c_, 1, 1, w=w.cv1)
        self.cv2 = TFConv(c_ * (len(k) + 1), c2, 1, 1, w=w.cv2)
        self.m = [keras.layers.MaxPool2D(pool_size=x, strides=1, padding='SAME') for x in k]

    def call(self, inputs):
        x = self.cv1(inputs)
        return self.cv2(tf.concat([x] + [m(x) for m in self.m], 3))


class TFDetect(keras.layers.Layer):
    def __init__(self, nc=80, anchors=(), ch=(), imgsz=(640, 640), w=None):  # detection layer
        super(TFDetect, self).__init__()
        self.stride = tf.convert_to_tensor(w.stride.numpy(), dtype=tf.float32)
        self.nc = nc  # number of classes
        self.no = nc + 5  # number of outputs per anchor
        self.nl = len(anchors)  # number of detection layers
        self.na = len(anchors[0]) // 2  # number of anchors
        self.grid = [tf.zeros(1)] * self.nl  # init grid
        self.anchors = tf.convert_to_tensor(w.anchors.numpy(), dtype=tf.float32)
        self.anchor_grid = tf.reshape(tf.convert_to_tensor(w.anchor_grid.numpy(), dtype=tf.float32),
                                      [self.nl, 1, -1, 1, 2])
        self.m = [TFConv2d(x, self.no * self.na, 1, w=w.m[i]) for i, x in enumerate(ch)]
        self.training = False  # set to False after building model
        self.imgsz = imgsz
        for i in range(self.nl):
            ny, nx = self.imgsz[0] // self.stride[i], self.imgsz[1] // self.stride[i]
            self.grid[i] = self._make_grid(nx, ny)

    def call(self, inputs):
        z = []  # inference output
        x = []
        for i in range(self.nl):
            x.append(self.m[i](inputs[i]))
            # x(bs,20,20,255) to x(bs,3,20,20,85)
            ny, nx = self.imgsz[0] // self.stride[i], self.imgsz[1] // self.stride[i]
            x[i] = tf.transpose(tf.reshape(x[i], [-1, ny * nx, self.na, self.no]), [0, 2, 1, 3])

            if not self.training:  # inference
                y = tf.sigmoid(x[i])
                xy = (y[..., 0:2] * 2. - 0.5 + self.grid[i]) * self.stride[i]  # xy
                wh = (y[..., 2:4] * 2) ** 2 * self.anchor_grid[i]
                # Normalize xywh to 0-1 to reduce calibration error
                xy /= tf.constant([[self.imgsz[1], self.imgsz[0]]], dtype=tf.float32)
                wh /= tf.constant([[self.imgsz[1], self.imgsz[0]]], dtype=tf.float32)
                y = tf.concat([xy, wh, y[..., 4:]], -1)
                z.append(tf.reshape(y, [-1, 3 * ny * nx, self.no]))

        return x if self.training else (tf.concat(z, 1), x)

    @staticmethod
    def _make_grid(nx=20, ny=20):
        # yv, xv = torch.meshgrid([torch.arange(ny), torch.arange(nx)])
        # return torch.stack((xv, yv), 2).view((1, 1, ny, nx, 2)).float()
        xv, yv = tf.meshgrid(tf.range(nx), tf.range(ny))
        return tf.cast(tf.reshape(tf.stack([xv, yv], 2), [1, 1, ny * nx, 2]), dtype=tf.float32)


class TFUpsample(keras.layers.Layer):
    def __init__(self, size, scale_factor, mode, w=None):  # warning: all arguments needed including 'w'
        super(TFUpsample, self).__init__()
        assert scale_factor == 2, "scale_factor must be 2"
        self.upsample = lambda x: tf.image.resize(x, (x.shape[1] * 2, x.shape[2] * 2), method=mode)
        # self.upsample = keras.layers.UpSampling2D(size=scale_factor, interpolation=mode)
        # with default arguments: align_corners=False, half_pixel_centers=False
        # self.upsample = lambda x: tf.raw_ops.ResizeNearestNeighbor(images=x,
        #                                                            size=(x.shape[1] * 2, x.shape[2] * 2))

    def call(self, inputs):
        return self.upsample(inputs)


class TFConcat(keras.layers.Layer):
    def __init__(self, dimension=1, w=None):
        super(TFConcat, self).__init__()
        assert dimension == 1, "convert only NCHW to NHWC concat"
        self.d = 3

    def call(self, inputs):
        return tf.concat(inputs, self.d)


def parse_model(d, ch, model, imgsz):  # model_dict, input_channels(3)
    LOGGER.info('\n%3s%18s%3s%10s  %-40s%-30s' % ('', 'from', 'n', 'params', 'module', 'arguments'))
    anchors, nc, gd, gw = d['anchors'], d['nc'], d['depth_multiple'], d['width_multiple']
    na = (len(anchors[0]) // 2) if isinstance(anchors, list) else anchors  # number of anchors
    no = na * (nc + 5)  # number of outputs = anchors * (classes + 5)

    layers, save, c2 = [], [], ch[-1]  # layers, savelist, ch out
    for i, (f, n, m, args) in enumerate(d['backbone'] + d['head']):  # from, number, module, args
        m_str = m
        m = eval(m) if isinstance(m, str) else m  # eval strings
        for j, a in enumerate(args):
            try:
                args[j] = eval(a) if isinstance(a, str) else a  # eval strings
            except:
                pass

        n = max(round(n * gd), 1) if n > 1 else n  # depth gain
        if m in [nn.Conv2d, Conv, Bottleneck, SPP, DWConv, MixConv2d, Focus, CrossConv, BottleneckCSP, C3]:
            c1, c2 = ch[f], args[0]
            c2 = make_divisible(c2 * gw, 8) if c2 != no else c2

            args = [c1, c2, *args[1:]]
            if m in [BottleneckCSP, C3]:
                args.insert(2, n)
                n = 1
        elif m is nn.BatchNorm2d:
            args = [ch[f]]
        elif m is Concat:
            c2 = sum([ch[-1 if x == -1 else x + 1] for x in f])
        elif m is Detect:
            args.append([ch[x + 1] for x in f])
            if isinstance(args[1], int):  # number of anchors
                args[1] = [list(range(args[1] * 2))] * len(f)
            args.append(imgsz)
        else:
            c2 = ch[f]

        tf_m = eval('TF' + m_str.replace('nn.', ''))
        m_ = keras.Sequential([tf_m(*args, w=model.model[i][j]) for j in range(n)]) if n > 1 \
            else tf_m(*args, w=model.model[i])  # module

        torch_m_ = nn.Sequential(*[m(*args) for _ in range(n)]) if n > 1 else m(*args)  # module
        t = str(m)[8:-2].replace('__main__.', '')  # module type
        np = sum([x.numel() for x in torch_m_.parameters()])  # number params
        m_.i, m_.f, m_.type, m_.np = i, f, t, np  # attach index, 'from' index, type, number params
        LOGGER.info('%3s%18s%3s%10.0f  %-40s%-30s' % (i, f, n, np, t, args))  # print
        save.extend(x % i for x in ([f] if isinstance(f, int) else f) if x != -1)  # append to savelist
        layers.append(m_)
        ch.append(c2)
    return keras.Sequential(layers), sorted(save)


class TFModel:
    def __init__(self, cfg='yolov5s.yaml', ch=3, nc=None, model=None, imgsz=(640, 640)):  # model, channels, classes
        super(TFModel, self).__init__()
        if isinstance(cfg, dict):
            self.yaml = cfg  # model dict
        else:  # is *.yaml
            import yaml  # for torch hub
            self.yaml_file = Path(cfg).name
            with open(cfg) as f:
                self.yaml = yaml.load(f, Loader=yaml.FullLoader)  # model dict

        # Define model
        if nc and nc != self.yaml['nc']:
            print('Overriding %s nc=%g with nc=%g' % (cfg, self.yaml['nc'], nc))
            self.yaml['nc'] = nc  # override yaml value
        self.model, self.savelist = parse_model(deepcopy(self.yaml), ch=[ch], model=model, imgsz=imgsz)

    def predict(self, inputs, tf_nms=False, agnostic_nms=False, topk_per_class=100, topk_all=100, iou_thres=0.45,
                conf_thres=0.25):
        y = []  # outputs
        x = inputs
        for i, m in enumerate(self.model.layers):
            if m.f != -1:  # if not from previous layer
                x = y[m.f] if isinstance(m.f, int) else [x if j == -1 else y[j] for j in m.f]  # from earlier layers

            x = m(x)  # run
            y.append(x if m.i in self.savelist else None)  # save output

        # Add TensorFlow NMS
        if tf_nms:
            boxes = self._xywh2xyxy(x[0][..., :4])
            probs = x[0][:, :, 4:5]
            classes = x[0][:, :, 5:]
            scores = probs * classes
            if agnostic_nms:
                nms = AgnosticNMS()((boxes, classes, scores), topk_all, iou_thres, conf_thres)
                return nms, x[1]
            else:
                boxes = tf.expand_dims(boxes, 2)
                nms = tf.image.combined_non_max_suppression(
                    boxes, scores, topk_per_class, topk_all, iou_thres, conf_thres, clip_boxes=False)
                return nms, x[1]

        return x[0]  # output only first tensor [1,6300,85] = [xywh, conf, class0, class1, ...]
        # x = x[0][0]  # [x(1,6300,85), ...] to x(6300,85)
        # xywh = x[..., :4]  # x(6300,4) boxes
        # conf = x[..., 4:5]  # x(6300,1) confidences
        # cls = tf.reshape(tf.cast(tf.argmax(x[..., 5:], axis=1), tf.float32), (-1, 1))  # x(6300,1)  classes
        # return tf.concat([conf, cls, xywh], 1)

    @staticmethod
    def _xywh2xyxy(xywh):
        # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
        x, y, w, h = tf.split(xywh, num_or_size_splits=4, axis=-1)
        return tf.concat([x - w / 2, y - h / 2, x + w / 2, y + h / 2], axis=-1)


class AgnosticNMS(keras.layers.Layer):
    # TF Agnostic NMS
    def call(self, input, topk_all, iou_thres, conf_thres):
        # wrap map_fn to avoid TypeSpec related error https://stackoverflow.com/a/65809989/3036450
        return tf.map_fn(lambda x: self._nms(x, topk_all, iou_thres, conf_thres), input,
                         fn_output_signature=(tf.float32, tf.float32, tf.float32, tf.int32),
                         name='agnostic_nms')

    @staticmethod
    def _nms(x, topk_all=100, iou_thres=0.45, conf_thres=0.25):  # agnostic NMS
        boxes, classes, scores = x
        class_inds = tf.cast(tf.argmax(classes, axis=-1), tf.float32)
        scores_inp = tf.reduce_max(scores, -1)
        selected_inds = tf.image.non_max_suppression(
            boxes, scores_inp, max_output_size=topk_all, iou_threshold=iou_thres, score_threshold=conf_thres)
        selected_boxes = tf.gather(boxes, selected_inds)
        padded_boxes = tf.pad(selected_boxes,
                              paddings=[[0, topk_all - tf.shape(selected_boxes)[0]], [0, 0]],
                              mode="CONSTANT", constant_values=0.0)
        selected_scores = tf.gather(scores_inp, selected_inds)
        padded_scores = tf.pad(selected_scores,
                               paddings=[[0, topk_all - tf.shape(selected_boxes)[0]]],
                               mode="CONSTANT", constant_values=-1.0)
        selected_classes = tf.gather(class_inds, selected_inds)
        padded_classes = tf.pad(selected_classes,
                                paddings=[[0, topk_all - tf.shape(selected_boxes)[0]]],
                                mode="CONSTANT", constant_values=-1.0)
        valid_detections = tf.shape(selected_inds)[0]
        return padded_boxes, padded_scores, padded_classes, valid_detections


def representative_dataset_gen(dataset, ncalib=100):
    # Representative dataset generator for use with converter.representative_dataset, returns a generator of np arrays
    for n, (path, img, im0s, vid_cap) in enumerate(dataset):
        input = np.transpose(img, [1, 2, 0])
        input = np.expand_dims(input, axis=0).astype(np.float32)
        input /= 255.0
        yield [input]
        if n >= ncalib:
            break


def run(weights=ROOT / 'yolov5s.pt',  # weights path
        imgsz=(640, 640),  # inference size h,w
        batch_size=1,  # batch size
        dynamic=False,  # dynamic batch size
        ):
    # PyTorch model
    im = torch.zeros((batch_size, 3, *imgsz))  # BCHW image
    model = attempt_load(weights, map_location=torch.device('cpu'), inplace=True, fuse=False)
    y = model(im)  # inference
    model.info()

    # TensorFlow model
    im = tf.zeros((batch_size, *imgsz, 3))  # BHWC image
    tf_model = TFModel(cfg=model.yaml, model=model, nc=model.nc, imgsz=imgsz)
    y = tf_model.predict(im)  # inference

    # Keras model
    im = keras.Input(shape=(*imgsz, 3), batch_size=None if dynamic else batch_size)
    keras_model = keras.Model(inputs=im, outputs=tf_model.predict(im))
    keras_model.summary()


def parse_opt():
    parser = argparse.ArgumentParser()
    parser.add_argument('--weights', type=str, default=ROOT / 'yolov5s.pt', help='weights path')
    parser.add_argument('--imgsz', '--img', '--img-size', nargs='+', type=int, default=[640], help='inference size h,w')
    parser.add_argument('--batch-size', type=int, default=1, help='batch size')
    parser.add_argument('--dynamic', action='store_true', help='dynamic batch size')
    opt = parser.parse_args()
    opt.imgsz *= 2 if len(opt.imgsz) == 1 else 1  # expand
    print_args(FILE.stem, opt)
    return opt


def main(opt):
    set_logging()
    run(**vars(opt))


if __name__ == "__main__":
    opt = parse_opt()
    main(opt)



from .data_sensor_mixin import _DataSensorMixin
from .sql_query_acquisition_mixin import _SqlQueryAcquisitionMixin
from .video_config import _VideoConfigMixin
from .video_file_mixin import _VideoFileMixin



import json
import os


from copy import deepcopy

from naeural_core import Logger
from naeural_core.bc import DefaultBlockEngine, BCct

def diff(str1, str2):
  offset = 40
  for i in range(1, len(str1)):
    if str1[:i] != str2[:i]:
      start = max(i-offset, 0)
      end = i + offset
      return str1[start:end], str2[start:end], i
  return None



if __name__ == '__main__':
  
  log = Logger("SGN2", base_folder='.', app_folder='_local_cache')
  
  LOCAL_NAME = 'e2'
  SUBFOLDER = 'payloads'
  
  bc_engine = DefaultBlockEngine(
    log=log,
    name=LOCAL_NAME,
    config= {
      "PEM_FILE"     : LOCAL_NAME + ".pem",
      "PASSWORD"     : None,      
      "PEM_LOCATION" : "data"
     },
  )
  
  folder = log.get_data_subfolder(SUBFOLDER)
  cases = list(set([x[7:].split('.')[0] for x in os.listdir(folder) if '_err' in x]))
  filtered = [x for x in cases if '1a9' in x]
  tests = cases #filtered[:1]
  for str_error in tests:
    fn_txt = os.path.join(folder,'_err-j-' + str_error + '.txt')
    with open(fn_txt, 'r') as fh:
      str_json = fh.read()
    fn_pkl = '_err-d-' + str_error + '.pkl'
    dct_data = log.load_pickle_from_data(fn_pkl, subfolder_path=SUBFOLDER)
    str_dct_data = log.safe_dumps_json(dct_data, replace_nan=False, ensure_ascii=False)
    msg_match = str_json == str_dct_data 
    if not msg_match:
      s1, s2, idx = diff(str_dct_data, str_json)
      log.P("Stringify match FAIL. Diff at {}: \n|{}|\n|{}|".format(idx, s1, s2), color='r')
    else:
      log.P("Stringify match: OK", color='g')     
    
    continue   
      
    assert dct_data == json.loads(str_json)
    msg_sender = dct_data[BCct.SENDER]
    msg_hash = dct_data[BCct.HASH]
    msg_sign = dct_data[BCct.SIGN]
    loc_sender = bc_engine.address
    prepared_data = log.replace_nan(dct_data, inplace=False)
    prepared_data == dct_data
    bdata, bin_hexdigest, hexdigest = bc_engine.compute_hash(
      dct_data, 
      return_all=True, 
      replace_nan=True,
    )
    color = 'g' if loc_sender == msg_sender else 'r'
    log.P("Msg sender: {}".format(msg_sender), color=color)
    log.P("Loc sender: {}".format(loc_sender), color=color)
    color = 'g' if hexdigest == msg_hash else 'r'
    log.P("Msg hash: {}".format(msg_hash), color=color)
    log.P("Loc hash: {}".format(hexdigest), color=color)

  



import json


from copy import deepcopy

from naeural_core import Logger
from naeural_core.bc import DefaultBlockEngine

def diff(str1, str2):
  offset = 40
  for i in range(1, len(str1)):
    if str1[:i] != str2[:i]:
      start = max(i-offset, 0)
      end = i + offset
      return str1[start:end], str2[start:end], i
  return None



if __name__ == '__main__':
  
  log = Logger("SGN", base_folder='.', app_folder='_local_cache')

  TESTS = {
    "gts-staging" : dict(
      be_recv_str =  """{"EE_FORMATTER":"cavi2","EE_HASH":"a4a72c99bdd561cc94612de1b4ebb806647637c50bc1da4da9c4c75f3ba7dc67","EE_MESSAGE_ID":"8f21aa82-1447-4dc8-821c-b4dc87657d95","EE_PAYLOAD_PATH":["gts-staging","LPR-1",null,null],"EE_SENDER":"0xai_AvDJaXULCUbtjvTFZikVGQM3SJoG7XXw3RK5U6LZHD7S","EE_SIGN":"MEUCIQCpNYZgw4Oc_RPGs4hkine8e8pwRHOC-4SlCKsHY5VaBQIgaonkJQE3VCZXoHvQNastqosr5Avs8A7yobbJZPPzPA8=","category":"","data":{"identifiers":{},"img":{"height":null,"id":null,"width":null},"specificValue":{},"time":null,"value":{}},"demoMode":false,"messageID":"8f21aa82-1447-4dc8-821c-b4dc87657d95","metadata":{"displayed":true,"ee_message_seq":6199,"ee_timezone":"UTC+2","ee_tz":"Europe/Bucharest","error_code":null,"info":"EE v3.28.264, Lib v9.8.71, info text: None","initiator_id":"cavi2-staging","module":"VideoStreamDataCapture","notification":"Video DCT 'LPR-1' successfully connected. Overall 1910 reconnects.","notification_code":null,"notification_tag":null,"notification_type":"NORMAL","sbCurrentMessage":"8f21aa82-1447-4dc8-821c-b4dc87657d95","sbTotalMessages":6199,"session_id":null,"stream_name":"LPR-1","timestamp":"2023-11-10 10:52:27.465499","video_stream_info":{"buffersize":0.0,"current_interval":null,"fps":23,"frame_count":429,"frame_current":814873,"frame_h":1080,"frame_w":1920}},"sender":{"hostId":"gts-staging","id":"edge-node","instanceId":"edge-node-v3.28.264"},"time":{"deviceTime":"","hostTime":"2023-11-10 10:52:28.492998","internetTime":""},"type":"notification","version":"3.28.264"}""",
      be_stringify = """{"EE_FORMATTER":"cavi2","EE_MESSAGE_ID":"8f21aa82-1447-4dc8-821c-b4dc87657d95","EE_PAYLOAD_PATH":["gts-staging","LPR-1",null,null],"category":"","data":{"identifiers":{},"img":{"height":null,"id":null,"width":null},"specificValue":{},"time":null,"value":{}},"demoMode":false,"messageID":"8f21aa82-1447-4dc8-821c-b4dc87657d95","metadata":{"displayed":true,"ee_message_seq":6199,"ee_timezone":"UTC+2","ee_tz":"Europe/Bucharest","error_code":null,"info":"EE v3.28.264, Lib v9.8.71, info text: None","initiator_id":"cavi2-staging","module":"VideoStreamDataCapture","notification":"Video DCT 'LPR-1' successfully connected. Overall 1910 reconnects.","notification_code":null,"notification_tag":null,"notification_type":"NORMAL","sbCurrentMessage":"8f21aa82-1447-4dc8-821c-b4dc87657d95","sbTotalMessages":6199,"session_id":null,"stream_name":"LPR-1","timestamp":"2023-11-10 10:52:27.465499","video_stream_info":{"buffersize":0,"current_interval":null,"fps":23,"frame_count":429,"frame_current":814873,"frame_h":1080,"frame_w":1920}},"sender":{"hostId":"gts-staging","id":"edge-node","instanceId":"edge-node-v3.28.264"},"time":{"deviceTime":"","hostTime":"2023-11-10 10:52:28.492998","internetTime":""},"type":"notification","version":"3.28.264"}""",
      be_gen_hash = "c0b40a298c63bd479a05517f4713c61f5bc898eb407fb7ac754a7de54c93d91b",
    ),
    
    "gts-test2" : dict(
      be_recv_str =  './core/xperimental/0xai_sign/bad-0.bin',
      be_stringify = './core/xperimental/0xai_sign/str1.json',
      be_gen_hash = "86a92270c175f8175d2cc65797d9bf85acc652575a426a9b52173fc7be50b936",
    ),
  }
  
  test_name = 'gts-test2'
  test = TESTS[test_name]


  be_recv_str = test['be_recv_str']
  if be_recv_str.startswith('./'):
    if be_recv_str.endswith('.bin'):
      with open(be_recv_str, 'rb') as f:
        be_recv_bin = f.read()
        be_recv_str = be_recv_bin.decode('utf-8')
        
    else:
      with open(be_recv_str, 'r', encoding="utf-8") as f:
        be_recv_str = f.read()
      
  be_stringify = test['be_stringify']
  if be_stringify.startswith('./'):
    with open(be_stringify, 'r', encoding="utf-8") as f:
      be_stringify = f.read()
      
  be_gen_hash = test['be_gen_hash']
  
  bc_engine = DefaultBlockEngine(
    log=log,
    name=test_name,
    config= {
      "PEM_FILE"     : test_name + ".pem",
      "PASSWORD"     : None,      
      "PEM_LOCATION" : "data"
     },
  )
  
  dct_to_send = json.loads(be_recv_str)
  be_rcv_hash = dct_to_send['EE_HASH']
  be_rcv_addr = dct_to_send['EE_SENDER']
  be_rcv_sign = dct_to_send['EE_SIGN'] 
  
  verify_msg = bc_engine.verify(dct_to_send)
  log.P("Verify msg: {}".format(verify_msg), color='b')

    
  d = deepcopy(dct_to_send)
  dct_local = {k:v for k,v in d.items() if k not in ['EE_SIGN','EE_SENDER','EE_HASH']}
  
  log.P("Loc addr:  {}".format(bc_engine.address), color='b')
  log.P("Msg addr:  {}".format(be_rcv_addr), color='b')
  # assert bc_engine.address == be_rcv_addr # only if the SK is available locally
  
  str_local_ = bc_engine._generate_data_for_hash(dct_to_send)
  bdata, bin_hash, local_hash = bc_engine.compute_hash(dct_local, return_all=True)
  str_local = bdata.decode('utf-8')
  assert str_local == str_local_, "Something is wrong with the json encoding/decoding"
  
  msg_match = str_local == be_stringify 
  if not msg_match:
    s1, s2, idx = diff(str_local, be_stringify)
    log.P("Stringify match FAIL. Diff at {}: \n|{}|\n|{}|".format(idx, s1, s2), color='r')
  else:
    log.P("Stringify match: OK", color='g')        

    log.P("Loc hash: {}".format(local_hash), color='b')
    log.P("Msg hash: {}".format(be_rcv_hash), color='b')
    log.P("BE hash:  {}".format(be_gen_hash), color='b')
    local_vs_be = local_hash == be_gen_hash
    if not local_vs_be:
      log.P("Hash generated from BE and locally match FAIL", color='r')
    else:
      log.P("Hash generated from BE and locally match OK", color='g')
        
      hash_check = local_hash == be_rcv_hash
      if not hash_check:
        log.P("Hash computed locally vs hash received on BE  check: FAIL", color='r')
      else:
        log.P("Hash computed locally vs hash received on BE check: OK", color='g')
        
        str_sign = bc_engine.sign(dct_local)
        # sanity check as this should be equal due to additions in dict
        # in dct_local the signature is the above
        log.P("Local sign: {}".format(dct_local['EE_SIGN']), color='b')
        log.P("BE rcv sgn: {}".format(dct_to_send['EE_SIGN']), color='b')
        # do not test local sign with remote as the local will be different due to algorithm  

          
        sign_reconstruct = str_sign == be_rcv_sign
        if sign_reconstruct:
          log.P("Sign JS reconstruct: OK", color='g') 
        else:
          log.P("Sign JS reconstruct: FAIL", color='r')
      
      


from naeural_core import DecentrAIObject
from naeural_core import Logger
from naeural_core.business.mixins_base import _DiskAPIMixin
import pandas as pd
from time import sleep
import numpy as np
import cv2

class A(DecentrAIObject, _DiskAPIMixin):

  def __init__(self, log, **kwargs):
    super(A, self).__init__(log=log, **kwargs)
    return

if __name__ == '__main__':
  l = Logger(lib_name='diskapi', base_folder='.', app_folder='_local_cache')
  a = A(log=l)


  ### define data for checking DiskAPI
  # dataframe
  col3 = []
  for i in range(3):
    col3.append(l.now_str(nice_print=True, short=False))
    sleep(1)

  data_for_df = {'col1' : [1,2,3], 'col2' : ['a', 'b', 'c'], 'col3' : col3}
  df = pd.DataFrame(data_for_df)

  # json
  dct = {'1' : 'abc', 'a' : [{2 : 'c'}, {3 : 'e'}], 'x' : False}

  # pickle
  arr = np.random.randint(0, 100, size=(100,100))

  # video
  lst_frames = []
  cap = cv2.VideoCapture('https://live.zootargoviste.ro:7001')
  nr_frames = 0
  H,W = None, None
  while nr_frames < 100:
    has_frame, frame = cap.read()
    if has_frame:
      H,W = frame.shape[:2]
      lst_frames.append(frame)
      nr_frames += 1

  cap.release()
  cv2.destroyAllWindows()

  #### use the DiskAPI

  # dataframe
  path_df1 = a.diskapi_save_dataframe_to_output(
    df=df, filename='diskapi_tests/dataframes/df1.csv'
  )

  path_df2 = a.diskapi_save_dataframe_to_output(
    df=df, filename='diskapi_tests/dataframes/df2.csv',
    compress=True
  )

  df1 = a.diskapi_load_dataframe_from_output(
    filename='diskapi_tests/dataframes/df1.csv',
    timestamps=['col3']
  )

  df2 = a.diskapi_load_dataframe_from_output(
    filename='diskapi_tests/dataframes/df2.csv',
    timestamps=['col3'],
    decompress=True
  )


  # json
  path_j1 = a.diskapi_save_json_to_output(
    dct=dct, filename='diskapi_tests/jsons/j1.json',
    indent=True
  )

  path_j2 = a.diskapi_save_json_to_output(
    dct=dct, filename='diskapi_tests/jsons/j2.json',
    indent=False
  )

  j2 = a.diskapi_load_json_from_output(filename='diskapi_tests/jsons/j2.json', verbose=True)

  # pickle
  path_p1 = a.diskapi_save_pickle_to_output(
    obj=arr, filename='diskapi_tests/pickles/p1.pickle',
    compress=False
  )

  path_p2 = a.diskapi_save_pickle_to_output(
    obj=arr, filename='diskapi_tests/pickles/p2.pickle',
    compress=True
  )

  p2 = a.diskapi_load_pickle_from_output(
    filename='diskapi_tests/pickles/p2.pickle',
    decompress=True
  )

  # video
  handle, path_v1 = a.diskapi_create_video_file_to_output(
    filename='diskapi_tests/videos/v1.mp4',
    fps=25, str_codec="mp4v", frame_size=(H,W)
  )

  for f in lst_frames:
    a.diskapi_write_video_frame(handle, f)













"""

This plugin uses external info (the hb data) in order to check the health of the system. 


"""

from naeural_core.business.base import BasePluginExecutor

__VER__ = '0.1.1.0'

_CONFIG = {
  **BasePluginExecutor.CONFIG,

  'ALLOW_EMPTY_INPUTS'            : True,
  
  "PROCESS_DELAY"                 : 5,
  
  "DISK_LOW_PRC"                  : 0.15,
  "MEM_LOW_PRC"                   : 0.15,
  
  
  "ALERT_RAISE_CONFIRMATION_TIME" : 1,
  "ALERT_LOWER_CONFIRMATION_TIME" : 2,
  "ALERT_DATA_COUNT"              : 2,
  "ALERT_RAISE_VALUE"             : 0.5,
  "ALERT_LOWER_VALUE"             : 0.4,
  "ALERT_MODE"                    : 'mean',
  "ALERT_MODE_LOWER"              : 'max',
  
  
  
  # debug stuff
  "LOG_INFO"            : False,

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES'],
  },
}


class SelfCheck01Plugin(BasePluginExecutor):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(SelfCheck01Plugin, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    return
  
  def on_command(self, data, **kwargs):
    self.P("Self check monitor on {} received: {}".format(self.eeid, data))
    return
  
  def check_all(self):
    is_issue = False
    dct_results = self.OrderedDict()
    dct_check_data = self.OrderedDict()
    for method_name, func in self._get_methods(self.__class__, include_parent=False):
      if method_name.startswith('_self_check_'):
        try:
          is_check_alert, msg, dct_alert_info = func(self)
          is_issue = is_issue or is_check_alert
          alerter_name = self.get_alert_name(method_name)
          check_name = self.get_check_name(method_name).upper()
          dct_alerter = self.alerter_status_dict(alerter=alerter_name)
          dct_alerter = {k.upper(): v for k, v in dct_alerter.items()}
          dct_check_data[check_name] = {
            "INFO" : msg,
            **dct_alert_info,
            **dct_alerter
          }
        except:
          err = "Programming error in self-check method '{}'.\n{}".format(method_name, self.trace_info())
          raise ValueError(err)
      #endif check method name
    #endfor check methods
    dct_results['CHECK_DATA'] = dct_check_data
    return is_issue, dct_results
  
  def get_result(self, avail : float, alert : float, check : str):
    return {
      "AVAIL_PRC": round(avail, 4),
      "ALERT_PRC": round(alert, 4),
      "CHECK": check,
    }    
    
  def get_check_name(self, func_name):
    return func_name.replace('_self_check_', '')
  
  def get_alert_name(self, func_name):
    return "check_" + self.get_check_name(func_name)
            
  def _self_check_memory(self):
    """
    Function for checking memory. As any `_self_check_` function it must return:
      - is_alert: True if alert is raised, False otherwise
      - msg: string with alert message
      - dct_result: dict with info for UI/BE
    """
    # start mandatory
    dct_result = self.OrderedDict()
    self_name = self.inspect.currentframe().f_code.co_name 
    alerter_name = self.get_alert_name(self_name)
    is_alert, msg = False, ""
    # end mandatory
    is_ok_mem = self.netmon.network_node_is_ok_available_memory_prc(
      addr=self.e2_addr, 
      min_prc_available=self.cfg_mem_low_prc,
    )      
    self.alerter_add_observation(not is_ok_mem, alerter=alerter_name) 
    
    if self.alerter_is_alert(alerter=alerter_name): #check if current alert exists
      mem_avail_prc = self.netmon.network_node_available_memory_prc(addr=self.e2_addr)      
      msg = "Memory low (free mem {:.1f}% < {:.1f}%) on node {} ({})".format(
          100 * mem_avail_prc, 100 * self.cfg_mem_low_prc, self.e2_addr, self.eeid,
      )
      is_alert = True
      dct_result = self.get_result(
        avail=mem_avail_prc, alert=self.cfg_mem_low_prc, check="Available RAM"
      )
    return is_alert, msg, dct_result
  
  
  def _self_check_disk(self):
    # start mandatory
    dct_result = self.OrderedDict()
    self_name = self.inspect.currentframe().f_code.co_name 
    alerter_name = self.get_alert_name(self_name)
    is_alert, msg = False, ""
    # end mandatory
    is_ok_disk = self.netmon.network_node_is_ok_available_disk_prc(
      addr=self.e2_addr, 
      min_prc_available=self.cfg_disk_low_prc
    )
    self.alerter_add_observation(not is_ok_disk, alerter=alerter_name) 
    
    if self.alerter_is_alert(alerter=alerter_name):  #check if current alert exists
      disk_avail_prc = self.netmon.network_node_available_disk_prc(addr=self.e2_addr)      
      msg = "Disk low (free disk {:.1f}% < {:.1f}%) on node {} ({})".format(
          100 * disk_avail_prc, 100 * self.cfg_mem_low_prc, self.e2_addr, self.eeid,
      ) 
      is_alert = True
      dct_result = self.get_result(
        avail=disk_avail_prc, alert=self.cfg_disk_low_prc, check="Available disk"
      )
    return is_alert, msg, dct_result
  
  
  def log_info(self, msg):
    if self.cfg_log_info:
      self.P(msg)
    return


  def _process(self):
    payload = None
    has_alerts = False

    self.log_info("Self check iteration...")    
    # run all chekers
    if not self.net_mon.network_node_info_available(addr=self.e2_addr):
      self.P("No hb info available for self at this moment {} ({})".format(self.e2_addr, self.eeid), color='r')
    else:
      has_alerts, dct_status = self.check_all()
      self.log_info("Self check iteration done. Has alerts: {}, Status:\n{}".format(
        has_alerts, self.json.dumps(dct_status, indent=4))
      )
      self.alerter_add_observation(int(has_alerts)) # default alerter
        
      if self.alerter_status_changed(): # check new raise or new lower
        # for this plugin only ALERTS should be used in UI/BE
        payload = self._create_payload(
          
          status="Self check alert raised" if has_alerts else "Self check alert lowered",
          **dct_status
        )
      #endif alert changed
    #endif info available    
    return payload